# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import io
import struct
import typing


class DES:
    """Creates a DES cipher.

    Creates a DES cipher class with the key initialised. This key must be 8 bytes in length. This only supports the
    ECB cipher mode as that is what is used in the LM hash calculation.

    Thanks to http://page.math.tu-berlin.de/~kant/teaching/hess/krypto-ws2006/des.htm for helping work through this.

    Args:
        key: The 8-byte key to use in the Cipher.
    """

    # First table used to derive the sub keys.
    _pc1 = [
        56,
        48,
        40,
        32,
        24,
        16,
        8,
        0,
        57,
        49,
        41,
        33,
        25,
        17,
        9,
        1,
        58,
        50,
        42,
        34,
        26,
        18,
        10,
        2,
        59,
        51,
        43,
        35,
        62,
        54,
        46,
        38,
        30,
        22,
        14,
        6,
        61,
        53,
        45,
        37,
        29,
        21,
        13,
        5,
        60,
        52,
        44,
        36,
        28,
        20,
        12,
        4,
        27,
        19,
        11,
        3,
    ]

    # Shifts the sub key from pc1 to calculate the 16 sub keys.
    _shift_indexes = [1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1]

    # Second table used to derive the sub keys.
    _pc2 = [
        13,
        16,
        10,
        23,
        0,
        4,
        2,
        27,
        14,
        5,
        20,
        9,
        22,
        18,
        11,
        3,
        25,
        7,
        15,
        6,
        26,
        19,
        12,
        1,
        40,
        51,
        30,
        36,
        46,
        54,
        29,
        39,
        50,
        44,
        32,
        47,
        43,
        48,
        38,
        55,
        33,
        52,
        45,
        41,
        49,
        35,
        28,
        31,
    ]

    # Initial permutation of the 64-bits of the message data.
    _ip = [
        57,
        49,
        41,
        33,
        25,
        17,
        9,
        1,
        59,
        51,
        43,
        35,
        27,
        19,
        11,
        3,
        61,
        53,
        45,
        37,
        29,
        21,
        13,
        5,
        63,
        55,
        47,
        39,
        31,
        23,
        15,
        7,
        56,
        48,
        40,
        32,
        24,
        16,
        8,
        0,
        58,
        50,
        42,
        34,
        26,
        18,
        10,
        2,
        60,
        52,
        44,
        36,
        28,
        20,
        12,
        4,
        62,
        54,
        46,
        38,
        30,
        22,
        14,
        6,
    ]

    # Used to expand the each initial permuted half into a 48-bit values.
    _e_bit_selection = [
        31,
        0,
        1,
        2,
        3,
        4,
        3,
        4,
        5,
        6,
        7,
        8,
        7,
        8,
        9,
        10,
        11,
        12,
        11,
        12,
        13,
        14,
        15,
        16,
        15,
        16,
        17,
        18,
        19,
        20,
        19,
        20,
        21,
        22,
        23,
        24,
        23,
        24,
        25,
        26,
        27,
        28,
        27,
        28,
        29,
        30,
        31,
        0,
    ]

    # List of boxes used in the encryption process.
    _s_boxes = [
        [
            14,
            4,
            13,
            1,
            2,
            15,
            11,
            8,
            3,
            10,
            6,
            12,
            5,
            9,
            0,
            7,
            0,
            15,
            7,
            4,
            14,
            2,
            13,
            1,
            10,
            6,
            12,
            11,
            9,
            5,
            3,
            8,
            4,
            1,
            14,
            8,
            13,
            6,
            2,
            11,
            15,
            12,
            9,
            7,
            3,
            10,
            5,
            0,
            15,
            12,
            8,
            2,
            4,
            9,
            1,
            7,
            5,
            11,
            3,
            14,
            10,
            0,
            6,
            13,
        ],
        [
            15,
            1,
            8,
            14,
            6,
            11,
            3,
            4,
            9,
            7,
            2,
            13,
            12,
            0,
            5,
            10,
            3,
            13,
            4,
            7,
            15,
            2,
            8,
            14,
            12,
            0,
            1,
            10,
            6,
            9,
            11,
            5,
            0,
            14,
            7,
            11,
            10,
            4,
            13,
            1,
            5,
            8,
            12,
            6,
            9,
            3,
            2,
            15,
            13,
            8,
            10,
            1,
            3,
            15,
            4,
            2,
            11,
            6,
            7,
            12,
            0,
            5,
            14,
            9,
        ],
        [
            10,
            0,
            9,
            14,
            6,
            3,
            15,
            5,
            1,
            13,
            12,
            7,
            11,
            4,
            2,
            8,
            13,
            7,
            0,
            9,
            3,
            4,
            6,
            10,
            2,
            8,
            5,
            14,
            12,
            11,
            15,
            1,
            13,
            6,
            4,
            9,
            8,
            15,
            3,
            0,
            11,
            1,
            2,
            12,
            5,
            10,
            14,
            7,
            1,
            10,
            13,
            0,
            6,
            9,
            8,
            7,
            4,
            15,
            14,
            3,
            11,
            5,
            2,
            12,
        ],
        [
            7,
            13,
            14,
            3,
            0,
            6,
            9,
            10,
            1,
            2,
            8,
            5,
            11,
            12,
            4,
            15,
            13,
            8,
            11,
            5,
            6,
            15,
            0,
            3,
            4,
            7,
            2,
            12,
            1,
            10,
            14,
            9,
            10,
            6,
            9,
            0,
            12,
            11,
            7,
            13,
            15,
            1,
            3,
            14,
            5,
            2,
            8,
            4,
            3,
            15,
            0,
            6,
            10,
            1,
            13,
            8,
            9,
            4,
            5,
            11,
            12,
            7,
            2,
            14,
        ],
        [
            2,
            12,
            4,
            1,
            7,
            10,
            11,
            6,
            8,
            5,
            3,
            15,
            13,
            0,
            14,
            9,
            14,
            11,
            2,
            12,
            4,
            7,
            13,
            1,
            5,
            0,
            15,
            10,
            3,
            9,
            8,
            6,
            4,
            2,
            1,
            11,
            10,
            13,
            7,
            8,
            15,
            9,
            12,
            5,
            6,
            3,
            0,
            14,
            11,
            8,
            12,
            7,
            1,
            14,
            2,
            13,
            6,
            15,
            0,
            9,
            10,
            4,
            5,
            3,
        ],
        [
            12,
            1,
            10,
            15,
            9,
            2,
            6,
            8,
            0,
            13,
            3,
            4,
            14,
            7,
            5,
            11,
            10,
            15,
            4,
            2,
            7,
            12,
            9,
            5,
            6,
            1,
            13,
            14,
            0,
            11,
            3,
            8,
            9,
            14,
            15,
            5,
            2,
            8,
            12,
            3,
            7,
            0,
            4,
            10,
            1,
            13,
            11,
            6,
            4,
            3,
            2,
            12,
            9,
            5,
            15,
            10,
            11,
            14,
            1,
            7,
            6,
            0,
            8,
            13,
        ],
        [
            4,
            11,
            2,
            14,
            15,
            0,
            8,
            13,
            3,
            12,
            9,
            7,
            5,
            10,
            6,
            1,
            13,
            0,
            11,
            7,
            4,
            9,
            1,
            10,
            14,
            3,
            5,
            12,
            2,
            15,
            8,
            6,
            1,
            4,
            11,
            13,
            12,
            3,
            7,
            14,
            10,
            15,
            6,
            8,
            0,
            5,
            9,
            2,
            6,
            11,
            13,
            8,
            1,
            4,
            10,
            7,
            9,
            5,
            0,
            15,
            14,
            2,
            3,
            12,
        ],
        [
            13,
            2,
            8,
            4,
            6,
            15,
            11,
            1,
            10,
            9,
            3,
            14,
            5,
            0,
            12,
            7,
            1,
            15,
            13,
            8,
            10,
            3,
            7,
            4,
            12,
            5,
            6,
            11,
            0,
            14,
            9,
            2,
            7,
            11,
            4,
            1,
            9,
            12,
            14,
            2,
            0,
            6,
            10,
            13,
            15,
            3,
            5,
            8,
            2,
            1,
            14,
            7,
            4,
            10,
            8,
            13,
            15,
            12,
            9,
            0,
            3,
            5,
            6,
            11,
        ],
    ]

    # Converts the s-box permutation one more time.
    _p = [
        15,
        6,
        19,
        20,
        28,
        11,
        27,
        16,
        0,
        14,
        22,
        25,
        4,
        17,
        30,
        9,
        1,
        7,
        23,
        13,
        31,
        26,
        2,
        8,
        18,
        12,
        29,
        5,
        21,
        10,
        3,
        24,
    ]

    # Final permutation of the message.
    _final_ip = [
        39,
        7,
        47,
        15,
        55,
        23,
        63,
        31,
        38,
        6,
        46,
        14,
        54,
        22,
        62,
        30,
        37,
        5,
        45,
        13,
        53,
        21,
        61,
        29,
        36,
        4,
        44,
        12,
        52,
        20,
        60,
        28,
        35,
        3,
        43,
        11,
        51,
        19,
        59,
        27,
        34,
        2,
        42,
        10,
        50,
        18,
        58,
        26,
        33,
        1,
        41,
        9,
        49,
        17,
        57,
        25,
        32,
        0,
        40,
        8,
        48,
        16,
        56,
        24,
    ]

    def __init__(self, key: bytes) -> None:
        if len(key) != 8:
            raise ValueError("DES encryption key should be 8 bytes in length")

        self.key = key
        self._subkeys = self._create_subkeys(self.key)

    def encrypt(self, data: bytes) -> bytes:
        """Encrypt a byte string.

        DES encrypts the data based on the key it was initialised with. The data will be right padded with b"\x00" if
        it is not a multiple of 8.

        Args:
            data: The byte string to encrypt.

        Returns:
            bytes: The encrypted byte string.
        """
        encrypted_data = io.BytesIO()

        for i in range(0, len(data), 8):
            block = data[i : i + 8].ljust(8, b"\x00")
            encrypted_data.write(self._encode_block(block))

        return encrypted_data.getvalue()

    def decrypt(self, data: bytes) -> bytes:
        """Decrypt a byte string.

        DES decrypts the data based on the key it was initialised with.

        Args:
            data: The byte string to decrypt

        Returns:
            bytes: The decrypted byte string.
        """
        decrypted_data = io.BytesIO()

        for i in range(0, len(data), 8):
            block = data[i : i + 8]
            if len(block) != 8:
                raise ValueError("DES decryption block must be a multiple of 8 bytes")

            decrypted_data.write(self._decode_block(block))

        return decrypted_data.getvalue()

    @staticmethod
    def key56_to_key64(key: bytes) -> bytes:
        """Convert 7 byte key to 8 bytes.

        This takes in an a bytes string of 7 bytes and converts it to a bytes
        string of 8 bytes with the odd parity bit being set to every 8 bits,

        For example

        b"\x01\x02\x03\x04\x05\x06\x07"
        00000001 00000010 00000011 00000100 00000101 00000110 00000111

        is converted to

        b"\x01\x80\x80\x61\x40\x29\x19\x0E"
        00000001 10000000 10000000 01100001 01000000 00101001 00011001 00001110

        https://crypto.stackexchange.com/questions/15799/des-with-actual-7-byte-key

        Args:
            key: The 7-byte sized key

        Returns:
            bytes: The expanded 8-byte key.
        """
        if len(key) != 7:
            raise ValueError("DES 7-byte key is not 7 bytes in length, actual: %d" % len(key))

        new_key = b""
        for i in range(0, 8):
            if i == 0:
                new_value = struct.unpack("B", key[i : i + 1])[0]
            elif i == 7:
                new_value = struct.unpack("B", key[6:7])[0]
                new_value = (new_value << 1) & 0xFF
            else:
                new_value = struct.unpack("B", key[i - 1 : i])[0]
                next_value = struct.unpack("B", key[i : i + 1])[0]
                new_value = ((new_value << (8 - i)) & 0xFF) | next_value >> i

            # clear the last bit so the count isn't off
            new_value = new_value & ~(1 << 0)

            # set the last bit if the number of set bits are even
            new_value = new_value | int(not DES.bit_count(new_value) & 0x1)
            new_key += struct.pack("B", new_value)

        return new_key

    @staticmethod
    def bit_count(i: int) -> int:
        # counts the number of bits that are 1 in the integer
        count = 0
        while i:
            i &= i - 1
            count += 1

        return count

    def _create_subkeys(self, key: bytes) -> typing.List[typing.List[int]]:
        # convert the key into a list of bits
        key_bits = self._get_bits(key)

        # reorder the bits based on the pc1 table
        pc1_bits = [key_bits[x] for x in self._pc1]

        # split the table into 2 and append to the first entry
        c = [pc1_bits[0:28]]
        d = [pc1_bits[28:56]]

        # now populate the remaining blocks by shifting the values
        for i, shift_index in enumerate(self._shift_indexes):
            c.append(self._shift_bits(c[i], shift_index))
            d.append(self._shift_bits(d[i], shift_index))

        subkeys: typing.List[typing.List[int]] = []
        for i in range(1, 17):
            cd = c[i] + d[i]
            subkey_bits = [cd[x] for x in self._pc2]
            subkeys.append(subkey_bits)

        return subkeys

    def _shift_bits(self, bits: typing.List[int], shifts: int) -> typing.List[int]:
        new_bits = [0] * 28
        for i in range(28):
            shift_index = i + shifts
            if shift_index >= 28:
                shift_index = shift_index - 28
            new_bits[i] = bits[shift_index]

        return new_bits

    def _get_bits(self, data: bytes) -> typing.List[int]:
        bits = []
        for i in range(len(data)):
            b = struct.unpack("B", data[i : i + 1])[0]
            bits.append(1 if b & 0x80 else 0)
            bits.append(1 if b & 0x40 else 0)
            bits.append(1 if b & 0x20 else 0)
            bits.append(1 if b & 0x10 else 0)
            bits.append(1 if b & 0x08 else 0)
            bits.append(1 if b & 0x04 else 0)
            bits.append(1 if b & 0x02 else 0)
            bits.append(1 if b & 0x01 else 0)
        return bits

    def _encode_block(self, block: bytes) -> bytes:
        block_bits = self._get_bits(block)
        lr = [block_bits[x] for x in self._ip]

        left = [lr[0:32]]
        right = [lr[32:64]]
        for i in range(16):
            computed_block = self._compute_block(right[i], self._subkeys[i])
            new_r = [int(computed_block[k] != left[i][k]) for k in range(32)]

            left.append(right[i])
            right.append(new_r)

        # apply the final permutation on the left and right bits backwards
        rl = right[16] + left[16]
        encrypted_bits = [rl[x] for x in self._final_ip]
        encrypted_bytes = b""
        for i in range(0, 64, 8):
            i_byte = int("".join([str(x) for x in encrypted_bits[i : i + 8]]), 2)
            encrypted_bytes += struct.pack("B", i_byte)

        return encrypted_bytes

    def _decode_block(self, block: bytes) -> bytes:
        block_bits = self._get_bits(block)
        rl = [0] * 64
        for i, idx in enumerate(self._final_ip):
            rl[idx] = block_bits[i]

        right: typing.List[typing.List[int]] = [[]] * 17
        left: typing.List[typing.List[int]] = [[]] * 17
        right[16] = rl[0:32]
        left[16] = rl[32:64]
        for i in range(15, -1, -1):
            computed_block = self._compute_block(left[i + 1], self._subkeys[i])
            new_l = [int(computed_block[k] != right[i + 1][k]) for k in range(32)]
            right[i] = left[i + 1]
            left[i] = new_l

        lr = left[0] + right[0]
        decrypted_bits = [0] * 64
        for i, idx in enumerate(self._ip):
            decrypted_bits[idx] = lr[i]

        decrypted_bytes = b""
        for i in range(0, 64, 8):
            i_byte = int("".join([str(x) for x in decrypted_bits[i : i + 8]]), 2)
            decrypted_bytes += struct.pack("B", i_byte)

        return decrypted_bytes

    def _compute_block(self, block: typing.List[int], key: typing.List[int]) -> typing.List[int]:
        expanded_block = [block[x] for x in self._e_bit_selection]
        new_block = [int(key[i] != expanded_block[i]) for i in range(48)]

        # calculate with the s-boxes
        s_box_perm = []
        s_box_iter = 0
        # now go through each block (8 groups of 6 bits) and run the s-boxes
        for i in range(0, 48, 6):
            current_block = new_block[i : i + 6]
            row_bits = [str(current_block[0]), str(current_block[-1])]
            column_bits = [str(x) for x in current_block[1:-1]]

            s_box_row = int("".join(row_bits), 2)
            s_box_column = int("".join(column_bits), 2)
            s_box_address = (s_box_row * 16) + s_box_column
            s_box_value = self._s_boxes[s_box_iter][s_box_address]
            s_box_iter += 1

            s_box_perm.append(1 if s_box_value & 0x8 else 0)
            s_box_perm.append(1 if s_box_value & 0x4 else 0)
            s_box_perm.append(1 if s_box_value & 0x2 else 0)
            s_box_perm.append(1 if s_box_value & 0x1 else 0)

        final_block = [s_box_perm[x] for x in self._p]
        return final_block



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime
import struct

import pytest

import spnego._ntlm_raw.messages as messages

from .._ntlm_raw import (
    TEST_NTLMV1_FLAGS,
    TEST_NTLMV2_FLAGS,
    TEST_SERVER_CHALLENGE,
    TEST_SERVER_NAME,
    TEST_USER,
    TEST_USER_DOM,
    TEST_WORKSTATION_NAME,
)


class UTC10(datetime.tzinfo):
    """Test UTC+10 timezone class."""

    def utcoffset(self, dt):
        return datetime.timedelta(hours=10)

    def tzname(self, dt):
        return "UTC+10"

    def dst(self, dt):
        return datetime.timedelta(hours=10)


def test_negotiate_flags_native_labels():
    actual = messages.NegotiateFlags.native_labels()

    assert isinstance(actual, dict)
    assert actual[messages.NegotiateFlags.key_56] == "NTLMSSP_NEGOTIATE_56"


def test_av_id_native_labels():
    actual = messages.AvId.native_labels()

    assert isinstance(actual, dict)
    assert actual[messages.AvId.channel_bindings] == "MSV_AV_CHANNEL_BINDINGS"


def test_av_flags_native_labels():
    actual = messages.AvFlags.native_labels()

    assert isinstance(actual, dict)
    assert actual[messages.AvFlags.mic] == "MIC_PROVIDED"


def test_message_type_native_labels():
    actual = messages.MessageType.native_labels()

    assert isinstance(actual, dict)
    assert actual[messages.MessageType.challenge] == "CHALLENGE_MESSAGE"


def test_negotiate_pack_defaults():
    negotiate = messages.Negotiate()

    assert negotiate.flags == 0
    assert negotiate.domain_name is None
    assert negotiate.workstation is None
    assert negotiate.version is None

    actual = negotiate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
    )

    negotiate.flags = 1

    actual = negotiate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
    )


def test_negotiate_pack_with_domain():
    negotiate = messages.Negotiate(domain_name="café", workstation="café")

    assert (
        negotiate.flags
        == messages.NegotiateFlags.oem_workstation_supplied | messages.NegotiateFlags.oem_domain_name_supplied
    )
    assert negotiate.domain_name == "café"
    assert negotiate.workstation == "café"
    assert negotiate.version is None

    actual = negotiate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x20\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x24\x00\x00\x00"
        b"\x63\x61\x66\xE9"
        b"\x63\x61\x66\xE9"
    )


def test_negotiate_pack_with_all_fields():
    negotiate = messages.Negotiate(domain_name="café", workstation="café", version=messages.Version(1, 1, 1))

    assert (
        negotiate.flags
        == messages.NegotiateFlags.oem_workstation_supplied
        | messages.NegotiateFlags.oem_domain_name_supplied
        | messages.NegotiateFlags.version
    )
    assert negotiate.domain_name == "café"
    assert negotiate.workstation == "café"
    assert negotiate.version == messages.Version(1, 1, 1)

    actual = negotiate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x02"
        b"\x04\x00"
        b"\x04\x00"
        b"\x28\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x2C\x00\x00\x00"
        b"\x01\x01\x01\x00\x00\x00\x00\x0F"
        b"\x63\x61\x66\xE9"
        b"\x63\x61\x66\xE9"
    )


def test_negotiate_pack_encoding():
    negotiate = messages.Negotiate(domain_name="café", workstation="café", encoding="utf-8")

    assert (
        negotiate.flags
        == messages.NegotiateFlags.oem_workstation_supplied | messages.NegotiateFlags.oem_domain_name_supplied
    )
    assert negotiate.domain_name == "café"
    assert negotiate.workstation == "café"
    assert negotiate.version is None

    actual = negotiate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x00"
        b"\x05\x00"
        b"\x05\x00"
        b"\x20\x00\x00\x00"
        b"\x05\x00"
        b"\x05\x00"
        b"\x25\x00\x00\x00"
        b"\x63\x61\x66\xC3\xA9"
        b"\x63\x61\x66\xC3\xA9"
    )


def test_negotiate_unpack():
    actual = messages.Negotiate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x20\x00\x00\x00"
    )

    assert actual.flags == 0
    assert actual.domain_name is None
    assert actual.workstation is None
    assert actual.version is None


def test_negotiate_unpack_with_domain():
    actual = messages.Negotiate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x20\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x24\x00\x00\x00"
        b"\x63\x61\x66\xE9"
        b"\x63\x61\x66\xE9"
    )

    assert (
        actual.flags
        == messages.NegotiateFlags.oem_workstation_supplied | messages.NegotiateFlags.oem_domain_name_supplied
    )
    assert actual.domain_name == "café"
    assert actual.workstation == "café"
    assert actual.version is None


def test_negotiate_unpack_with_all_fields():
    actual = messages.Negotiate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x02"
        b"\x04\x00"
        b"\x04\x00"
        b"\x28\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x2C\x00\x00\x00"
        b"\x01\x01\x01\x00\x00\x00\x00\x0F"
        b"\x63\x61\x66\xE9"
        b"\x63\x61\x66\xE9"
    )

    assert (
        actual.flags
        == messages.NegotiateFlags.oem_workstation_supplied
        | messages.NegotiateFlags.oem_domain_name_supplied
        | messages.NegotiateFlags.version
    )

    assert actual.domain_name == "café"
    assert actual.workstation == "café"
    assert actual.version == messages.Version(1, 1, 1)


def test_negotiate_unpack_encoding():
    actual = messages.Negotiate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x00"
        b"\x05\x00"
        b"\x05\x00"
        b"\x20\x00\x00\x00"
        b"\x05\x00"
        b"\x05\x00"
        b"\x25\x00\x00\x00"
        b"\x63\x61\x66\xC3\xA9"
        b"\x63\x61\x66\xC3\xA9",
        encoding="utf-8",
    )

    assert (
        actual.flags
        == messages.NegotiateFlags.oem_workstation_supplied | messages.NegotiateFlags.oem_domain_name_supplied
    )
    assert actual.domain_name == "café"
    assert actual.workstation == "café"
    assert actual.version is None


def test_negotiate_unpack_invalid_encoding():
    # While the cars are invalid UTF-8 chars we don't want that to raise an exception for the Negotiate msg.
    actual = messages.Negotiate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x30\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x20\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x24\x00\x00\x00"
        b"\x63\x61\x66\xE9"
        b"\x63\x61\x66\xE9",
        encoding="utf-8",
    )

    assert (
        actual.flags
        == messages.NegotiateFlags.oem_workstation_supplied | messages.NegotiateFlags.oem_domain_name_supplied
    )
    assert actual.domain_name == "caf�"
    assert actual.workstation == "caf�"
    assert actual.version is None


def test_negotiate_invalid_size():
    with pytest.raises(ValueError, match="Invalid NTLM Negotiate raw byte length"):
        messages.Negotiate.unpack(b"NTLMSSP\x00\x01\x00\x00\x00")


def test_negotiate_unpack_invalid_msg():
    with pytest.raises(ValueError, match="Input message was not a NTLM Negotiate message"):
        messages.Negotiate.unpack(
            b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
            b"\x02\x00\x00\x00"
            b"\x0C\x00"
            b"\x0C\x00"
            b"\x38\x00\x00\x00"
            b"\x33\x82\x02\xE2"
            b"\x01\x23\x45\x67\x89\xAB\xCD\xEF"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00"
            b"\x00\x00"
            b"\x00\x00\x00\x00"
            b"\x06\x00\x70\x17\x00\x00\x00\x0F"
            b"\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        )


def test_challenge_pack():
    challenge = messages.Challenge()

    assert challenge.flags == 0
    assert challenge.server_challenge == b"\x00" * 8
    assert challenge.target_info is None
    assert challenge.target_name is None
    assert challenge.version is None

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x30\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x30\x00\x00\x00"
    )


def test_challenge_pack_target_name():
    challenge = messages.Challenge(
        flags=messages.NegotiateFlags.unicode, server_challenge=b"\x11" * 8, target_name="café"
    )

    assert challenge.flags == messages.NegotiateFlags.unicode | messages.NegotiateFlags.request_target
    assert challenge.server_challenge == b"\x11" * 8
    assert challenge.target_name == "café"

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x30\x00\x00\x00"
        b"\x05\x00\x00\x00"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x38\x00\x00\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00"
    )


def test_challenge_pack_target_name_oem():
    challenge = messages.Challenge(flags=messages.NegotiateFlags.oem, server_challenge=b"\x11" * 8, target_name="café")

    assert challenge.flags == messages.NegotiateFlags.oem | messages.NegotiateFlags.request_target
    assert challenge.server_challenge == b"\x11" * 8
    assert challenge.target_name == "café"

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x30\x00\x00\x00"
        b"\x06\x00\x00\x00"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x34\x00\x00\x00"
        b"\x63\x61\x66\xE9"
    )


def test_challenge_pack_target_info():
    ti = messages.TargetInfo()
    ti[messages.AvId.dns_computer_name] = "café"
    # Even with the OEM encoding flag, the target info should still be utf-16-le.
    challenge = messages.Challenge(flags=messages.NegotiateFlags.oem, server_challenge=b"\x11" * 8, target_info=ti)

    assert challenge.flags == messages.NegotiateFlags.oem | messages.NegotiateFlags.target_info
    assert challenge.server_challenge == b"\x11" * 8
    assert challenge.target_name is None
    assert challenge.target_info is not None
    assert len(challenge.target_info) == 2
    assert challenge.target_info[messages.AvId.dns_computer_name] == "café"
    assert challenge.target_info[messages.AvId.eol] == b""
    assert challenge.version is None

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x30\x00\x00\x00"
        b"\x02\x00\x80\x00"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x30\x00\x00\x00"
        b"\x03\x00\x08\x00\x63\x00\x61\x00\x66\x00\xE9\x00"
        b"\x00\x00\x00\x00"
    )


def test_challenge_pack_all_fields():
    ti = messages.TargetInfo()
    ti[messages.AvId.dns_computer_name] = "café"
    challenge = messages.Challenge(
        flags=messages.NegotiateFlags.unicode,
        server_challenge=b"\x11" * 8,
        target_name="café",
        target_info=ti,
        version=messages.Version(2, 2, 2),
    )

    assert (
        challenge.flags
        == messages.NegotiateFlags.target_info
        | messages.NegotiateFlags.request_target
        | messages.NegotiateFlags.version
        | messages.NegotiateFlags.unicode
    )
    assert challenge.server_challenge == b"\x11" * 8
    assert challenge.target_name == "café"
    assert challenge.target_info is not None
    assert len(challenge.target_info) == 2
    assert challenge.target_info[messages.AvId.dns_computer_name] == "café"
    assert challenge.target_info[messages.AvId.eol] == b""
    assert challenge.version == messages.Version(2, 2, 2)

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x38\x00\x00\x00"
        b"\x05\x00\x80\x02"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x40\x00\x00\x00"
        b"\x02\x02\x02\x00\x00\x00\x00\x0F"
        b"\x63\x00\x61\x00\x66\x00\xE9"
        b"\x00\x03\x00\x08\x00\x63\x00\x61\x00\x66\x00\xE9\x00"
        b"\x00\x00\x00\x00"
    )


def test_challenge_set_fields():
    challenge = messages.Challenge()

    challenge.flags = 10
    challenge.server_challenge = b"\xFF" * 8

    actual = challenge.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x30\x00\x00\x00"
        b"\x0A\x00\x00\x00"
        b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x30\x00\x00\x00"
    )


def test_challenge_invalid_server_challenge_length():
    expected = "NTLM Challenge ServerChallenge must be 8 bytes long"
    with pytest.raises(ValueError, match=expected):
        messages.Challenge(server_challenge=b"\x00")

    challenge = messages.Challenge()
    with pytest.raises(ValueError, match=expected):
        challenge.server_challenge = b"\x08"


def test_challenge_invalid_size():
    with pytest.raises(ValueError, match="Invalid NTLM Challenge raw byte length"):
        messages.Challenge.unpack(b"NTLMSSP\x00\x02\x00\x00\x00")


def test_challenge_unpack_ntlmv1():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/ee78f3ad-ae29-4de1-96a0-fe46e64b6e31
    actual = messages.Challenge.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x38\x00\x00\x00"
        b"\x33\x82\x02\xE2"
        b"\x01\x23\x45\x67\x89\xAB\xCD\xEF"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x06\x00\x70\x17\x00\x00\x00\x0F"
        b"\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
    )

    assert actual.target_name == TEST_SERVER_NAME
    assert actual.flags == TEST_NTLMV1_FLAGS
    assert actual.server_challenge == TEST_SERVER_CHALLENGE
    assert actual.target_info is None
    assert actual.version == messages.Version(6, 0, 6000)


def test_challenge_unpack_ntlmv2():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/bc612491-fb0b-4829-91bc-7c6b95ff67fe
    actual = messages.Challenge.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x38\x00\x00\x00"
        b"\x33\x82\x8A\xE2"
        b"\x01\x23\x45\x67\x89\xAB\xCD\xEF"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x24\x00"
        b"\x24\x00"
        b"\x44\x00\x00\x00"
        b"\x06\x00\x70\x17\x00\x00\x00\x0F"
        b"\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
    )

    assert actual.target_name == TEST_SERVER_NAME
    assert actual.flags == TEST_NTLMV2_FLAGS
    assert actual.server_challenge == TEST_SERVER_CHALLENGE
    assert actual.target_info is not None
    assert len(actual.target_info) == 3
    assert actual.target_info[messages.AvId.nb_domain_name] == TEST_USER_DOM
    assert actual.target_info[messages.AvId.nb_computer_name] == TEST_SERVER_NAME
    assert actual.target_info[messages.AvId.eol] == b""
    assert actual.version == messages.Version(6, 0, 6000)


def test_challenge_unpack_encoding():
    actual = messages.Challenge.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x02\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x30\x00\x00\x00"
        b"\x06\x00\x00\x00"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x34\x00\x00\x00"
        b"\x63\x61\x66\xE9"
    )

    assert actual.target_name == "café"
    assert actual.flags == messages.NegotiateFlags.oem | messages.NegotiateFlags.request_target
    assert actual.server_challenge == b"\x11" * 8
    assert actual.target_info is None
    assert actual.version is None


def test_challenge_unpack_invalid_msg():
    with pytest.raises(ValueError, match="Input message was not a NTLM Challenge message"):
        messages.Challenge.unpack(
            b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
            b"\x01\x00\x00\x00"
            b"\x00\x00\x00\x00"
            b"\x00\x00"
            b"\x00\x00"
            b"\x20\x00\x00\x00"
            b"\x00\x00"
            b"\x00\x00"
            b"\x20\x00\x00\x00"
        )


def test_authenticate_pack():
    authenticate = messages.Authenticate()

    assert authenticate.lm_challenge_response is None
    assert authenticate.nt_challenge_response is None
    assert authenticate.domain_name is None
    assert authenticate.user_name is None
    assert authenticate.workstation is None
    assert authenticate.encrypted_random_session_key is None
    assert authenticate.flags == 0
    assert authenticate.version is None
    assert authenticate.mic == b"\x00" * 16

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )


def test_authenticate_pack_domain_version():
    authenticate = messages.Authenticate(
        flags=messages.NegotiateFlags.unicode, domain_name="café", version=messages.Version(1, 1, 1)
    )

    assert authenticate.lm_challenge_response is None
    assert authenticate.nt_challenge_response is None
    assert authenticate.domain_name == "café"
    assert authenticate.user_name is None
    assert authenticate.workstation is None
    assert authenticate.encrypted_random_session_key is None
    assert authenticate.flags == messages.NegotiateFlags.version | messages.NegotiateFlags.unicode
    assert authenticate.mic == b"\x00" * 16
    assert authenticate.version == messages.Version(1, 1, 1)

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x58\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x58\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x58\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x60\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x60\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x60\x00\x00\x00"
        b"\x01\x00\x00\x02"
        b"\x01\x01\x01\x00\x00\x00\x00\x0F"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00"
    )


def test_authenticate_pack_oem_encoding():
    authenticate = messages.Authenticate(
        flags=messages.NegotiateFlags.oem, domain_name="café", version=messages.Version(1, 1, 1)
    )

    assert authenticate.lm_challenge_response is None
    assert authenticate.nt_challenge_response is None
    assert authenticate.domain_name == "café"
    assert authenticate.user_name is None
    assert authenticate.workstation is None
    assert authenticate.encrypted_random_session_key is None
    assert authenticate.flags == messages.NegotiateFlags.version | messages.NegotiateFlags.oem
    assert authenticate.mic == b"\x00" * 16
    assert authenticate.version == messages.Version(1, 1, 1)

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x58\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x58\x00\x00\x00"
        b"\x04\x00"
        b"\x04\x00"
        b"\x58\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x5C\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x5C\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x5C\x00\x00\x00"
        b"\x02\x00\x00\x02"
        b"\x01\x01\x01\x00\x00\x00\x00\x0F"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x63\x61\x66\xE9"
    )


def test_authenticate_set_flags():
    authenticate = messages.Authenticate()
    authenticate.flags = 1

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )


def test_authenticate_set_mic():
    authenticate = messages.Authenticate()
    authenticate.mic = b"\xFF" * 16

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
    )


def test_authenticate_set_encrypted_session_key():
    authenticate = messages.Authenticate(encrypted_session_key=b"\x01")

    actual = authenticate.pack()

    assert (
        actual == b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x01\x00"
        b"\x01\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00\x00\x40"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x01"
    )


def test_authenticate_invalid_mic_length():
    expected = "NTLM Authenticate MIC must be 16 bytes long"

    with pytest.raises(ValueError, match=expected):
        messages.Authenticate(mic=b"\x00")

    authenticate = messages.Authenticate()
    with pytest.raises(ValueError, match=expected):
        authenticate.mic = b"\x00"


def test_authenticate_set_mic_not_present():
    authenticate = messages.Authenticate.unpack(b"NTLMSSP\x00\x03\x00\x00\x00" + (b"\x00" * 52))

    assert authenticate.mic is None

    with pytest.raises(ValueError, match="Cannot set MIC on an Authenticate message with no MIC present"):
        authenticate.mic = b"\x11" * 16


def test_authenticate_unpack_empty():
    actual = messages.Authenticate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\x50\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )

    assert actual.lm_challenge_response is None
    assert actual.nt_challenge_response is None
    assert actual.domain_name is None
    assert actual.user_name is None
    assert actual.workstation is None
    assert actual.encrypted_random_session_key is None
    assert actual.flags == 0
    assert actual.version is None
    assert actual.mic == b"\x00" * 16


def test_authenticate_unpack_ntlmv1():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/ee78f3ad-ae29-4de1-96a0-fe46e64b6e31
    actual = messages.Authenticate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x6C\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x84\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x48\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x54\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x5C\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x9C\x00\x00\x00"
        b"\x35\x82\x80\xE2"
        b"\x05\x01\x28\x0A\x00\x00\x00\x0F"
        b"\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x55\x00\x73\x00\x65\x00\x72\x00"
        b"\x43\x00\x4F\x00\x4D\x00\x50\x00\x55\x00\x54\x00\x45\x00\x52\x00"
        b"\x98\xDE\xF7\xB8\x7F\x88\xAA\x5D\xAF\xE2\xDF\x77\x96\x88\xA1\x72"
        b"\xDE\xF1\x1C\x7D\x5C\xCD\xEF\x13"
        b"\x67\xC4\x30\x11\xF3\x02\x98\xA2\xAD\x35\xEC\xE6\x4F\x16\x33\x1C"
        b"\x44\xBD\xBE\xD9\x27\x84\x1F\x94"
        b"\x51\x88\x22\xB1\xB3\xF3\x50\xC8\x95\x86\x82\xEC\xBB\x3E\x3C\xB7"
    )

    assert (
        actual.lm_challenge_response == b"\x98\xDE\xF7\xB8\x7F\x88\xAA\x5D"
        b"\xAF\xE2\xDF\x77\x96\x88\xA1\x72"
        b"\xDE\xF1\x1C\x7D\x5C\xCD\xEF\x13"
    )
    assert (
        actual.nt_challenge_response == b"\x67\xC4\x30\x11\xF3\x02\x98\xA2"
        b"\xAD\x35\xEC\xE6\x4F\x16\x33\x1C"
        b"\x44\xBD\xBE\xD9\x27\x84\x1F\x94"
    )
    assert actual.domain_name == TEST_USER_DOM
    assert actual.user_name == TEST_USER
    assert actual.workstation == TEST_WORKSTATION_NAME
    assert (
        actual.encrypted_random_session_key == b"\x51\x88\x22\xB1\xB3\xF3\x50\xC8" b"\x95\x86\x82\xEC\xBB\x3E\x3C\xB7"
    )
    assert actual.flags == 3800072757
    assert actual.mic is None
    assert actual.version == messages.Version(5, 1, 2600)


def test_authenticate_unpack_ntlmv2():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/bc612491-fb0b-4829-91bc-7c6b95ff67fe
    actual = messages.Authenticate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x6C\x00\x00\x00"
        b"\x54\x00"
        b"\x54\x00"
        b"\x84\x00\x00\x00"
        b"\x0C\x00"
        b"\x0C\x00"
        b"\x48\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\x54\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\x5C\x00\x00\x00"
        b"\x10\x00"
        b"\x10\x00"
        b"\xD8\x00\x00\x00"
        b"\x35\x82\x88\xE2"
        b"\x05\x01\x28\x0A\x00\x00\x00\x0F"
        b"\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x55\x00\x73\x00\x65\x00\x72\x00"
        b"\x43\x00\x4F\x00\x4D\x00\x50\x00\x55\x00\x54\x00\x45\x00\x52\x00"
        b"\x86\xC3\x50\x97\xAC\x9C\xEC\x10\x25\x54\x76\x4A\x57\xCC\xCC\x19"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x68\xCD\x0A\xB8\x51\xE5\x1C\x96\xAA\xBC\x92\x7B\xEB\xEF\x6A\x1C"
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x00\x00\x00\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\xC5\xDA\xD2\x54\x4F\xC9\x79\x90\x94\xCE\x1C\xE9\x0B\xC9\xD0\x3E"
    )

    assert (
        actual.lm_challenge_response == b"\x86\xC3\x50\x97\xAC\x9C\xEC\x10"
        b"\x25\x54\x76\x4A\x57\xCC\xCC\x19"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
    )
    assert (
        actual.nt_challenge_response == b"\x68\xCD\x0A\xB8\x51\xE5\x1C\x96"
        b"\xAA\xBC\x92\x7B\xEB\xEF\x6A\x1C"
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x00\x00\x00\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )
    assert actual.domain_name == TEST_USER_DOM
    assert actual.user_name == TEST_USER
    assert actual.workstation == TEST_WORKSTATION_NAME
    assert (
        actual.encrypted_random_session_key == b"\xC5\xDA\xD2\x54\x4F\xC9\x79\x90" b"\x94\xCE\x1C\xE9\x0B\xC9\xD0\x3E"
    )
    assert actual.flags == 3800597045
    assert actual.mic is None
    assert actual.version == messages.Version(5, 1, 2600)


def test_authenticate_unpack_mic_no_version():
    actual = messages.Authenticate.unpack(
        b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
        b"\x03\x00\x00\x00"
        b"\x18\x00"
        b"\x18\x00"
        b"\x50\x00\x00\x00"
        b"\x54\x00"
        b"\x54\x00"
        b"\x68\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\xBC\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\xC4\x00\x00\x00"
        b"\x08\x00"
        b"\x08\x00"
        b"\xCC\x00\x00\x00"
        b"\x00\x00"
        b"\x00\x00"
        b"\xD4\x00\x00\x00"
        b"\x31\x82\x8A\xE2"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x86\xC3\x50\x97\xAC\x9C\xEC\x10"
        b"\x25\x54\x76\x4A\x57\xCC\xCC\x19"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x68\xCD\x0A\xB8\x51\xE5\x1C\x96"
        b"\xAA\xBC\x92\x7B\xEB\xEF\x6A\x1C"
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x00\x00\x00\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00"
        b"\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00"
        b"\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00"
    )

    assert (
        actual.lm_challenge_response == b"\x86\xC3\x50\x97\xAC\x9C\xEC\x10"
        b"\x25\x54\x76\x4A\x57\xCC\xCC\x19"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
    )
    assert (
        actual.nt_challenge_response == b"\x68\xCD\x0A\xB8\x51\xE5\x1C\x96"
        b"\xAA\xBC\x92\x7B\xEB\xEF\x6A\x1C"
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\xAA\xAA\xAA\xAA\xAA\xAA\xAA\xAA"
        b"\x00\x00\x00\x00"
        b"\x02\x00\x0C\x00\x44\x00\x6F\x00\x6D\x00\x61\x00\x69\x00\x6E\x00"
        b"\x01\x00\x0C\x00\x53\x00\x65\x00\x72\x00\x76\x00\x65\x00\x72\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )
    assert actual.domain_name == "café"
    assert actual.user_name == "café"
    assert actual.workstation == "café"
    assert actual.encrypted_random_session_key is None
    assert actual.flags == 3800728113
    assert actual.mic == b"\x11" * 16
    assert actual.version is None


def test_authenticate_invalid_size():
    with pytest.raises(ValueError, match="Invalid NTLM Authenticate raw byte length"):
        messages.Authenticate.unpack(b"NTLMSSP\x00\x03\x00\x00\x00")


def test_authenticate_unpack_invalid_msg():
    with pytest.raises(ValueError, match="Input message was not a NTLM Authenticate message"):
        messages.Authenticate.unpack(
            b"\x4E\x54\x4C\x4D\x53\x53\x50\x00"
            b"\x01\x00\x00\x00"
            b"\x00\x00\x00\x00"
            b"\x00\x00"
            b"\x00\x00"
            b"\x20\x00\x00\x00"
            b"\x00\x00"
            b"\x00\x00"
            b"\x20\x00\x00\x00"
        )


def test_filetime_pack():
    filetime = messages.FileTime(1970, 1, 1, 0, 0, 0)
    actual = filetime.pack()

    assert actual == b"\x00\x80\x3E\xD5\xDE\xB1\x9D\x01"


def test_filetime_unpack():
    actual = messages.FileTime.unpack(b"\x00\x80\x3E\xD5\xDE\xB1\x9D\x01")

    assert isinstance(actual, messages.FileTime)
    assert str(actual) == "1970-01-01T00:00:00Z"
    assert actual.year == 1970
    assert actual.month == 1
    assert actual.day == 1
    assert actual.hour == 0
    assert actual.minute == 0
    assert actual.second == 0
    assert actual.microsecond == 0
    assert actual.nanosecond == 0


def test_filetime_from_datetime_nanoseconds():
    filetime = messages.FileTime.from_datetime(datetime.datetime(1970, 1, 1, 0, 0, 0), ns=500)
    actual = filetime.pack()

    assert str(filetime) == "1970-01-01T00:00:00.0000005Z"
    assert filetime.nanosecond == 500

    assert actual == b"\x05\x80\x3E\xD5\xDE\xB1\x9D\x01"


def test_filetime_now():
    current = messages.FileTime.from_datetime(datetime.datetime.now())
    now = messages.FileTime.now()

    current_int = struct.unpack("<Q", current.pack())[0]
    now_int = struct.unpack("<Q", now.pack())[0]

    assert now_int >= current_int


def test_filetime_with_timezone():
    filetime = messages.FileTime(1970, 1, 1, 10, 0, 0, tzinfo=UTC10())

    assert str(filetime) == "1970-01-01T10:00:00+10:00"

    actual = filetime.pack()  # Should be the same as EPOCH in UTC as FILETIME.
    assert actual == b"\x00\x80\x3E\xD5\xDE\xB1\x9D\x01"


def test_nt_challenge_pack():
    challenge = messages.NTClientChallengeV2()
    assert challenge.resp_type == 1
    assert challenge.hi_resp_type == 1
    assert isinstance(challenge.time_stamp, messages.FileTime)
    assert challenge.challenge_from_client == b"\x00" * 8
    assert len(challenge.av_pairs) == 1

    challenge.time_stamp = messages.FileTime.unpack(b"\x00\x00\x00\x00\x00\x00\x00\x00")

    actual = challenge.pack()

    assert (
        actual == b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )


def test_nt_challenge_create():
    ft = messages.FileTime.unpack(b"\x11" * 8)
    av = messages.TargetInfo()
    av[messages.AvId.dns_domain_name] = "test"

    actual = messages.NTClientChallengeV2(time_stamp=ft, client_challenge=b"\x11" * 8, av_pairs=av)

    assert actual.resp_type == 1
    assert actual.hi_resp_type == 1
    assert actual.time_stamp.pack() == b"\x11" * 8
    assert actual.challenge_from_client == b"\x11" * 8
    assert len(actual.av_pairs) == 2
    assert actual.av_pairs[messages.AvId.dns_domain_name] == "test"

    assert (
        actual.pack() == b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x11\x11\x11\x11\x11\x11\x11\x11"
        b"\x00\x00\x00\x00"
        b"\x04\x00\x08\x00"
        b"\x74\x00\x65\x00\x73\x00\x74\x00"
        b"\x00\x00\x00\x00"
    )


def test_nt_challenge_unpack():
    challenge = messages.NTClientChallengeV2.unpack(
        b"\x01"
        b"\x01"
        b"\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )

    assert challenge.resp_type == 1
    assert challenge.hi_resp_type == 1
    assert isinstance(challenge.time_stamp, messages.FileTime)
    assert challenge.time_stamp.pack() == b"\x00\x00\x00\x00\x00\x00\x00\x00"
    assert challenge.challenge_from_client == b"\x00\x00\x00\x00\x00\x00\x00\x00"
    assert isinstance(challenge.av_pairs, messages.TargetInfo)
    assert challenge.av_pairs.pack() == b"\x00\x00\x00\x00"


def test_nt_challenge_unpack_invalid_size():
    with pytest.raises(ValueError, match="Invalid NTClientChallengeV2 raw byte length"):
        messages.NTClientChallengeV2.unpack(b"\x00")


def test_nt_challenge_resp_type():
    challenge = messages.NTClientChallengeV2()
    assert challenge.resp_type == 1
    challenge.resp_type = 2
    assert challenge.resp_type == 2
    assert challenge.pack()[:1] == b"\x02"


def test_nt_challenge_hi_resp_type():
    challenge = messages.NTClientChallengeV2()
    assert challenge.hi_resp_type == 1
    challenge.hi_resp_type = 2
    assert challenge.hi_resp_type == 2
    assert challenge.pack()[1:2] == b"\x02"


def test_nt_challenge_client_challenge():
    challenge = messages.NTClientChallengeV2()
    assert challenge.challenge_from_client == b"\x00\x00\x00\x00\x00\x00\x00\x00"
    challenge.challenge_from_client = b"\xFF" * 8
    assert challenge.challenge_from_client == b"\xFF" * 8
    assert challenge.pack()[16:24] == b"\xFF" * 8


def test_nt_challenge_client_challenge_bad_length():
    expected = "NTClientChallengeV2 ChallengeFromClient must be 8 bytes long"

    with pytest.raises(ValueError, match=expected):
        messages.NTClientChallengeV2(client_challenge=b"\x00")

    challenge = messages.NTClientChallengeV2()
    with pytest.raises(ValueError, match=expected):
        challenge.challenge_from_client = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00"


def test_target_info_pack():
    target_info = messages.TargetInfo()
    target_info[messages.AvId.timestamp] = b"\x00\x00\x00\x00\x00\x00\x00\x00"
    target_info[messages.AvId.single_host] = b"\x00" * 48
    target_info[messages.AvId.dns_computer_name] = "caf\u00e9-host"
    target_info[messages.AvId.eol] = b""
    target_info[messages.AvId.channel_bindings] = b"\xFF" * 16
    target_info[messages.AvId.flags] = messages.AvFlags.constrained

    actual = target_info.pack()

    assert (
        actual == b"\x07\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x08\x00\x30\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x03\x00\x12\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00\x2D\x00\x68\x00\x6F\x00\x73\x00"
        b"\x74\x00"
        b"\x0A\x00\x10\x00"
        b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        b"\x06\x00\x04\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )


def test_target_info_unpack():
    actual = messages.TargetInfo.unpack(
        b"\x07\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x08\x00\x30\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x03\x00\x12\x00"
        b"\x63\x00\x61\x00\x66\x00\xE9\x00\x2D\x00\x68\x00\x6F\x00\x73\x00"
        b"\x74\x00"
        b"\x0A\x00\x10\x00"
        b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"
        b"\x06\x00\x04\x00"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )

    assert isinstance(actual, messages.TargetInfo)
    assert len(actual) == 6
    assert actual[messages.AvId.timestamp] == messages.FileTime.unpack(b"\x00" * 8)
    assert actual[messages.AvId.single_host] == messages.SingleHost.unpack(b"\x00" * 48)
    assert actual[messages.AvId.dns_computer_name] == "caf\u00e9-host"
    assert actual[messages.AvId.channel_bindings] == b"\xFF" * 16
    assert actual[messages.AvId.flags] == messages.AvFlags.constrained
    assert actual[messages.AvId.eol] == b""


def test_single_host_pack():
    single_host = messages.SingleHost(size=4, z4=0, custom_data=b"\x01" * 8, machine_id=b"\x02" * 32)
    actual = single_host.pack()

    assert (
        actual == b"\x04\x00\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01"
        b"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02"
        b"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02"
    )


def test_single_host_defaults():
    actual = messages.SingleHost()

    assert actual.size == 0
    assert actual.z4 == 0
    assert actual.custom_data == b"\x00" * 8
    assert actual.machine_id == b"\x00" * 32


def test_single_host_invalid_size():
    with pytest.raises(ValueError, match="SingleHost bytes must have a length of 48"):
        messages.SingleHost.unpack(b_data=b"\x00")


def test_single_host_invalid_custom_data_size():
    single_host = messages.SingleHost()

    with pytest.raises(ValueError, match="custom_data length must be 8 bytes long"):
        single_host.custom_data = b"\x00"


def test_single_host_invalid_machine_id_size():
    single_host = messages.SingleHost()

    with pytest.raises(ValueError, match="machine_id length must be 32 bytes long"):
        single_host.machine_id = b"\x00"


def test_single_host_unpack():
    actual = messages.SingleHost.unpack(
        b"\x04\x00\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01"
        b"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02"
        b"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02"
    )

    assert isinstance(actual, messages.SingleHost)
    assert actual.size == 4
    assert actual.z4 == 0
    assert actual.custom_data == b"\x01" * 8
    assert actual.machine_id == b"\x02" * 32


def test_single_host_eq():
    assert messages.SingleHost.unpack(b"\x00" * 48) == b"\x00" * 48
    assert messages.SingleHost.unpack(b"\x00" * 48) != b"\x11" * 48
    assert messages.SingleHost.unpack(b"\x00" * 48) != 1
    assert messages.SingleHost.unpack(b"\x00" * 48) == messages.SingleHost.unpack(b"\x00" * 48)


def test_version_pack():
    version = messages.Version(major=1, minor=2, build=3, revision=4)
    assert version.major == 1
    assert version.minor == 2
    assert version.build == 3
    assert version.revision == 4
    assert len(version) == 8

    version.major = 1
    version.minor = 1
    version.build = 1
    version.revision = 10
    actual = version.pack()

    assert actual == b"\x01\x01\x01\x00\x00\x00\x00\x0A"


def test_version_defaults():
    version = messages.Version()

    assert version.major == 0
    assert version.minor == 0
    assert version.build == 0
    assert version.reserved == b"\x00\x00\x00"
    assert version.revision == 15
    assert len(version) == 8


def test_version_unpack():
    actual = messages.Version.unpack(b"\x01\x01\x01\x00\x00\x00\x00\x0F")

    assert isinstance(actual, messages.Version)
    assert str(actual) == "1.1.1.15"
    assert repr(actual) == "<spnego._ntlm_raw.messages.Version 1.1.1.15>"
    assert actual.major == 1
    assert actual.minor == 1
    assert actual.build == 1
    assert actual.reserved == b"\x00\x00\x00"
    assert actual.revision == 15
    assert len(actual) == 8


def test_version_unpack_incorrect_length():
    with pytest.raises(ValueError, match="Version bytes must have a length of 8"):
        messages.Version.unpack(b"\x00")


def test_version_get_current():
    actual = messages.Version.get_current()

    assert isinstance(actual, messages.Version)
    assert len(actual) == 8
    assert len(actual.pack()) == 8


@pytest.mark.parametrize(
    "version, major, minor, build",
    [
        ("0.1.0.dev1", 0, 1, 0),
        ("0.1.0b1", 0, 1, 0),
        ("0.1.10b1", 0, 1, 10),
        ("0.1.0a1", 0, 1, 0),
        ("0.1.0", 0, 1, 0),
        ("0", 0, 0, 0),
        ("1", 1, 0, 0),
        ("1.2", 1, 2, 0),
        ("1..", 1, 0, 0),
    ],
)
def test_version_get_current_formats(version, major, minor, build, monkeypatch):
    monkeypatch.setattr(messages, "pyspnego_version", version)

    actual = messages.Version.get_current()
    assert actual.major == major
    assert actual.minor == minor
    assert actual.build == build


def test_version_eq():
    assert messages.Version.get_current() != messages.Version()
    assert messages.Version() == b"\x00" * 7 + b"\x0F"
    assert messages.Version() != b"\x11" * 8
    assert messages.Version() != 1
    assert messages.Version.get_current() == messages.Version.get_current()



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os

import pytest

from spnego._ntlm_raw.des import DES


def test_des_invalid_key_size():
    with pytest.raises(ValueError, match="DES encryption key should be 8 bytes in length") as exc:
        DES(b"\x01\x02")


def test_expand_56bit_key():
    expected = b"\x01\x80\x80\x61\x40\x29\x19\x0E"
    actual = DES.key56_to_key64(b"\x01\x02\x03\x04\x05\x06\x07")
    assert actual == expected

    expected = b"\x51\x20\x54\x6b\x34\xba\x3d\xa4"
    actual = DES.key56_to_key64(b"PASSWOR")
    assert actual == expected

    expected = b"\x45\x01\x01\x01\x01\x01\x01\x01"
    actual = DES.key56_to_key64(b"D\x00\x00\x00\x00\x00\x00")
    assert actual == expected


def test_encrypt_block():
    with pytest.raises(ValueError, match="DES 7-byte key is not 7 bytes in length, actual: 1") as exc:
        DES.key56_to_key64(b"\x00")


def test_encrypt():
    des = DES(b"PASSWORD")
    expected = b"\x15\x14\x4f\x75\x8c\x83\xd0\x34"
    actual = des.encrypt(b"abcdefgh")
    assert actual == expected


def test_encrypt_large_bytes_padding():
    des = DES(b"PASSWORD")
    expected = (
        b"\x15\x14\x4f\x75\x8c\x83\xd0\x34"
        b"\xe6\x19\x4b\xe6\x51\x22\xe3\x91"
        b"\x68\x02\xa9\xd8\x6c\x04\x1f\x2d"
        b"\x95\x1d\xc7\x12\xe4\x7c\x79\x04"
    )
    actual = des.encrypt(b"abcdefghijklmnopqrstuvwxyz")
    assert actual == expected


def test_decrypt():
    des = DES(b"PASSWORD")
    expected = b"abcdefgh"
    actual = des.decrypt(b"\x15\x14\x4f\x75\x8c\x83\xd0\x34")
    assert actual == expected


def test_decrypt_large_bytes_padding():
    des = DES(b"PASSWORD")
    expected = b"abcdefghijklmnopqrstuvwxyz\x00\x00\x00\x00\x00\x00"
    data = (
        b"\x15\x14\x4f\x75\x8c\x83\xd0\x34"
        b"\xe6\x19\x4b\xe6\x51\x22\xe3\x91"
        b"\x68\x02\xa9\xd8\x6c\x04\x1f\x2d"
        b"\x95\x1d\xc7\x12\xe4\x7c\x79\x04"
    )
    actual = des.decrypt(data)
    assert actual == expected


def test_decrypt_fail_invalid_size():
    with pytest.raises(ValueError, match="DES decryption block must be a multiple of 8 bytes") as exc:
        des = DES(b"PASSWORD")
        des.decrypt(b"\x01\x02\x03\x04")


def test_encrypt_decrypt_multiple_keys():
    # run random tests with random keys and data
    for i in range(512):
        des = DES(os.urandom(8))
        data = os.urandom(16)
        enc_data = des.encrypt(data)
        dec_data = des.decrypt(enc_data)
        assert enc_data != data
        assert dec_data == data



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re

import pytest

import spnego.exceptions as exceptions

GSSError = exceptions.GSSError
WinError = exceptions.WinError


@pytest.mark.parametrize(
    "option, expected",
    [
        (
            exceptions.NegotiateOptions.negotiate_kerberos,
            "The Python gssapi library is not installed so Kerberos cannot be negotiated.",
        ),
        (
            exceptions.NegotiateOptions.wrapping_iov,
            "The system is missing the GSSAPI IOV extension headers or CredSSP is being requested, "
            "cannot utilize wrap_iov and unwrap_iov",
        ),
        (
            exceptions.NegotiateOptions.wrapping_winrm,
            "The system is missing the GSSAPI IOV extension headers required for WinRM encryption with Kerberos.",
        ),
        (exceptions.NegotiateOptions.session_key, "The protocol selected does not support getting the session key."),
    ],
)
def test_feature_missing_error(option, expected):
    err = exceptions.FeatureMissingError(option)
    assert str(err) == expected
    assert err.message == expected


def test_spnego_error_no_code_fail():
    expected = "SpnegoError requires either an error_code or base_error"
    with pytest.raises(ValueError, match=re.escape(expected)):
        exceptions.SpnegoError()


def test_spnego_error_invalid_base_error_fail():
    expected = "base_error of type 'Exception' is not supported, must be a gssapi.exceptions.GSSError or WindowsError"
    with pytest.raises(ValueError, match=re.escape(expected)):
        exceptions.SpnegoError(base_error=Exception())


def test_spnego_error_unknown_error():
    actual = exceptions.SpnegoError(1024)

    assert isinstance(actual, exceptions.SpnegoError)
    assert actual.message == "SpnegoError (1024): Unknown error code"


def test_invalid_token_error():
    actual = exceptions.SpnegoError(exceptions.ErrorCode.invalid_token)

    assert isinstance(actual, exceptions.InvalidTokenError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.invalid_token
    assert actual.base_error is None
    assert actual.message == "SpnegoError (9): A token was invalid, or the logon was denied"


def test_invalid_token_error_with_context():
    actual = exceptions.InvalidTokenError(context_msg="Context")

    assert isinstance(actual, exceptions.InvalidTokenError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.invalid_token
    assert actual.base_error is None
    assert actual.message == "SpnegoError (9): A token was invalid, or the logon was denied, Context: Context"


@pytest.mark.skipif(GSSError == Exception, reason="Need a GSSError to test this out")
def test_invalid_token_from_gssapi():
    base_error = GSSError(589824, 0)

    actual = exceptions.SpnegoError(base_error=base_error)
    assert isinstance(actual, exceptions.InvalidTokenError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.invalid_token
    assert actual.base_error == base_error
    assert actual.message.startswith("SpnegoError (9): Major (589824)")


@pytest.mark.skipif(WinError == Exception, reason="Need a WindowsError to test this out")
def test_invalid_token_from_sspi():
    base_error = WinError("Error")
    setattr(base_error, "winerror", -2146893048)

    actual = exceptions.SpnegoError(base_error=base_error)
    assert isinstance(actual, exceptions.InvalidTokenError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.invalid_token
    assert actual.base_error == base_error
    assert actual.message.startswith("SpnegoError (9): ")


@pytest.mark.skipif(WinError == Exception, reason="Need a WindowsError to test this out")
def test_invalid_token_from_sspi_logon_denied():
    base_error = WinError("Error")
    setattr(base_error, "winerror", -2146893044)

    actual = exceptions.SpnegoError(base_error=base_error)
    assert isinstance(actual, exceptions.InvalidTokenError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.invalid_token
    assert actual.base_error == base_error
    assert actual.message.startswith("SpnegoError (9): ")


def test_operation_not_available_error():
    actual = exceptions.SpnegoError(exceptions.ErrorCode.unavailable)

    assert isinstance(actual, exceptions.OperationNotAvailableError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.unavailable
    assert actual.base_error is None
    assert actual.message == "SpnegoError (16): Operation not supported or available"


def test_operation_not_available_error_with_context():
    actual = exceptions.OperationNotAvailableError(context_msg="Context")

    assert isinstance(actual, exceptions.OperationNotAvailableError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.unavailable
    assert actual.base_error is None
    assert actual.message == "SpnegoError (16): Operation not supported or available, Context: Context"


@pytest.mark.skipif(GSSError == Exception, reason="Need a GSSError to test this out")
def test_operation_not_available_from_gssapi():
    base_error = GSSError(1048576, 0)

    actual = exceptions.SpnegoError(base_error=base_error)
    assert isinstance(actual, exceptions.OperationNotAvailableError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.unavailable
    assert actual.base_error == base_error
    assert actual.message.startswith("SpnegoError (16): Major (1048576): ")


@pytest.mark.skipif(WinError == Exception, reason="Need a WindowsError to test this out")
def test_operation_not_available_from_sspi():
    base_error = WinError("Error")
    setattr(base_error, "winerror", -2146893054)

    actual = exceptions.SpnegoError(base_error=base_error)
    assert isinstance(actual, exceptions.OperationNotAvailableError)
    assert actual.ERROR_CODE == exceptions.ErrorCode.unavailable
    assert actual.base_error == base_error
    assert actual.message.startswith("SpnegoError (16): ")



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import socket

import pytest

from spnego._text import to_bytes, to_text

HAS_K5TEST = True
try:
    from k5test.realm import K5Realm
except Exception:  # Can fail with more than ImportError on some systems
    HAS_K5TEST = False

HAS_SSPI = True
try:
    import win32net
    import win32netcon
except ImportError:
    HAS_SSPI = False


def get_data(name: str) -> bytes:
    with open(os.path.join(os.path.dirname(__file__), "data", name), mode="rb") as fd:
        return fd.read()


@pytest.fixture()
def ntlm_cred(tmpdir, monkeypatch):
    cleanup = None
    try:
        # Use unicode credentials to test out edge cases when dealing with non-ascii chars.
        username = "ÜseӜ"
        password = "Pӓ$sw0r̈d"

        if HAS_SSPI:
            domain = to_text(socket.gethostname())

            # Can only test this out with Windows due to issue with gss-ntlmssp when dealing with surrogate pairs.
            # https://github.com/gssapi/gss-ntlmssp/issues/20
            clef = to_text(b"\xF0\x9D\x84\x9E")
            username += clef
            password += clef

            buff = {
                "name": username,
                "password": password,
                "priv": win32netcon.USER_PRIV_USER,
                "comment": "Test account for pypsnego tests",
                "flags": win32netcon.UF_NORMAL_ACCOUNT,
            }
            try:
                win32net.NetUserAdd(None, 1, buff)
            except win32net.error as err:
                if err.winerror != 2224:  # Account already exists
                    raise

            def cleanup() -> None:
                win32net.NetUserDel(None, username)

        else:
            domain = "Dȫm̈Ąiᴞ"

        tmp_creds = os.path.join(to_text(tmpdir), "pÿspᴞӛgӫ TÈ$" ".creds")
        with open(tmp_creds, mode="wb") as fd:
            fd.write(to_bytes("%s:%s:%s" % (domain, username, password)))

        monkeypatch.setenv("NTLM_USER_FILE", to_text(tmp_creds))

        yield "%s\\%s" % (domain, username), password

    finally:
        if cleanup is not None:
            cleanup()


@pytest.fixture()
def kerb_cred(monkeypatch):
    if not HAS_K5TEST:
        pytest.skip("Cannot create Kerberos credential without k5test being installed")

    realm = K5Realm()
    try:
        for k, v in realm.env.items():
            monkeypatch.setenv(to_text(k), to_text(v))

        yield realm

    finally:
        realm.stop()
        del realm



# -*- coding: utf-8 -*-
# Copyright: (c) 2022, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)
import typing

import pytest

import spnego._tls_struct as tls


@pytest.mark.parametrize(
    "enum_cls, value, expected",
    [
        (tls.TlsProtocolVersion, 0x0A, "Unknown TLS Protocol Version 0x000A"),
        (tls.TlsContentType, 0x63, "Unknown TLS Content Type 0x63"),
        (tls.TlsHandshakeMessageType, 0xFF, "Unknown Handshake Message Type 0xFF"),
        (tls.TlsCipherSuite, 0xFF01, "Unknown Cipher Suite 0xFF01"),
        (tls.TlsCompressionMethod, 0x05, "Unknown Compression Method 0x05"),
        (tls.TlsExtensionType, 0x63, "Unknown Extension Type 0x0063"),
        (tls.TlsServerNameType, 2, "Unknown Server Name Type 0x02"),
        (tls.TlsECPointFormat, 3, "Unknown EC Point Format 0x03"),
        (tls.TlsSupportedGroup, 0x1254, "Unknown Supported Group 0x1254"),
        (tls.TlsSignatureScheme, 0x0001, "Unknown Signature Scheme 0x0001"),
        (tls.TlsPskKeyExchangeMode, 0x44, "Unknown PSK Key Exchange Mode 0x44"),
        (tls.TlsECCurveType, 0x55, "Unknown EC Curve Type 0x55"),
        (tls.TlsClientCertificateType, 0x70, "Unknown Client Certificate Type 0x70"),
        (tls.DistinguishedNameType, "1.2.3.4", "Unknown DN OID Type 1.2.3.4"),
    ],
)
def test_tls_enum_missing_member(enum_cls, value, expected):
    actual = enum_cls(value)
    assert isinstance(actual, enum_cls)
    assert actual.name == expected
    assert actual.value == value


def test_tls_enum_fail_non_int():
    with pytest.raises(ValueError, match="'fail' is not a valid TlsProtocolVersion"):
        tls.TlsProtocolVersion("fail")  # type: ignore[arg-type] # Testing a failure here



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime

import spnego._kerberos as kerb
import spnego._spnego as sp
from spnego._asn1 import (
    TagClass,
    TypeTagNumber,
    pack_asn1,
    pack_asn1_bit_string,
    pack_asn1_general_string,
    pack_asn1_integer,
    pack_asn1_octet_string,
    pack_asn1_sequence,
)
from spnego._context import GSSMech

from .conftest import get_data


def test_parse_enum_unknown():
    actual = kerb.parse_enum(10, enum_type=kerb.KerberosAPOptions)
    assert actual == "UNKNOWN (10)"


def test_parse_flags_unknown():
    actual = kerb.parse_flags(10, enum_type=kerb.KerberosAPOptions)
    assert actual["raw"] == 10
    assert actual["flags"] == ["UNKNOWN (10)"]


def test_kerberos_ap_options_native_labels():
    actual = kerb.KerberosAPOptions.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosAPOptions.mutual_required] == "mutual-required"


def test_kerberos_kdc_options_native_labels():
    actual = kerb.KerberosKDCOptions.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosKDCOptions.forwardable] == "forwardable"


def test_kerberos_encryption_type_native_labels():
    actual = kerb.KerberosEncryptionType.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosEncryptionType.des_cbc_crc] == "DES_CBC_CRC"


def test_kerberos_error_code_native_labels():
    actual = kerb.KerberosErrorCode.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosErrorCode.none] == "KDC_ERR_NONE"


def test_kerberos_message_type_native_labels():
    actual = kerb.KerberosMessageType.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosMessageType.as_req] == "AS-REQ"


def test_kerberos_pa_data_type_native_labels():
    actual = kerb.KerberosPADataType.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosPADataType.tgs_req] == "PA-TGS-REQ"


def test_kerberos_principal_name_type_native_labels():
    actual = kerb.KerberosPrincipalNameType.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosPrincipalNameType.principal] == "NT-PRINCIPAL"


def test_kerberos_host_address_type_native_labels():
    actual = kerb.KerberosHostAddressType.native_labels()

    assert isinstance(actual, dict)
    assert actual[kerb.KerberosHostAddressType.ipv4] == "IPv4"


def test_unpack_unknown_krb():
    sequence = pack_asn1_sequence(
        [
            pack_asn1(TagClass.context_specific, True, 0, pack_asn1_integer(5)),
            pack_asn1(TagClass.context_specific, True, 1, pack_asn1_integer(0)),
        ]
    )
    actual = sp.unpack_token(sequence, unwrap=True)

    assert isinstance(actual, kerb.KerberosV5Msg)
    assert actual.PVNO == 5
    assert isinstance(actual.sequence, dict)

    assert actual.sequence[0].tag_class == TagClass.universal
    assert not actual.sequence[0].constructed
    assert actual.sequence[0].tag_number == TypeTagNumber.integer
    assert actual.sequence[0].b_data == b"\x05"

    assert actual.sequence[1].tag_class == TagClass.universal
    assert not actual.sequence[1].constructed
    assert actual.sequence[1].tag_number == TypeTagNumber.integer
    assert actual.sequence[1].b_data == b"\x00"


def test_unpack_krb_as_req():
    data = get_data("krb_as_req")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, kerb.KrbAsReq)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.as_req
    assert isinstance(actual.padata, list)
    assert len(actual.padata) == 2

    assert actual.padata[0].data_type == kerb.KerberosPADataType.enc_timestamp
    assert (
        actual.padata[0].b_value == b"\x30\x41\xA0\x03\x02\x01\x12\xA2\x3A\x04\x38\x07\x40\x46\x03\xA8"
        b"\x69\xC9\x31\x76\xE2\x8E\xDA\xD1\x34\xCE\x7F\xC4\xC8\x73\x58\x0D"
        b"\xF4\x61\x1C\x85\x5F\x43\xF6\xAA\x9E\x48\xE2\xF0\x8C\xC2\x88\x70"
        b"\xAA\xBC\xF0\xF7\xF2\xD4\xA2\xC2\xE3\x53\xDE\x81\xF7\x30\x2F\xAF"
        b"\x7C\x85\x12"
    )
    pa1_val = actual.padata[0].value
    assert isinstance(pa1_val, kerb.EncryptedData)
    assert pa1_val.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert pa1_val.kvno is None
    assert (
        pa1_val.cipher == b"\x07\x40\x46\x03\xA8\x69\xC9\x31\x76\xE2\x8E\xDA\xD1\x34\xCE\x7F"
        b"\xC4\xC8\x73\x58\x0D\xF4\x61\x1C\x85\x5F\x43\xF6\xAA\x9E\x48\xE2"
        b"\xF0\x8C\xC2\x88\x70\xAA\xBC\xF0\xF7\xF2\xD4\xA2\xC2\xE3\x53\xDE"
        b"\x81\xF7\x30\x2F\xAF\x7C\x85\x12"
    )

    assert actual.padata[1].data_type == 149
    assert actual.padata[1].b_value == b""
    assert actual.padata[1].value == b""

    assert isinstance(actual.req_body, kerb.KdcReqBody)
    assert actual.req_body.additional_tickets is None
    assert actual.req_body.addresses is None
    assert actual.req_body.cname == kerb.PrincipalName(kerb.KerberosPrincipalNameType.principal, [b"vagrant-domain"])
    assert actual.req_body.enc_authorization_data is None
    assert actual.req_body.etype == [
        kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96,
        kerb.KerberosEncryptionType.aes128_cts_hmac_sha1_96,
        kerb.KerberosEncryptionType.des3_cbc_sha1,
        kerb.KerberosEncryptionType.rc4_hmac,
    ]
    assert actual.req_body.kdc_options == 1073741824
    assert actual.req_body.nonce == 734266074
    assert actual.req_body.postdated_from is None
    assert actual.req_body.postdated_till == datetime.datetime(2020, 6, 14, 7, 4, 20, tzinfo=datetime.timezone.utc)
    assert actual.req_body.realm == b"DOMAIN.LOCAL"
    assert actual.req_body.rtime is None
    assert actual.req_body.sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.srv_inst, [b"krbtgt", b"DOMAIN.LOCAL"]
    )

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert isinstance(actual, dict)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "AS-REQ (10)"
    assert isinstance(actual["padata"], list)
    assert len(actual["padata"]) == 2
    assert actual["padata"][0]["padata-type"] == "PA-ENC-TIMESTAMP (2)"
    assert actual["padata"][0]["padata-value"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["padata"][0]["padata-value"]["kvno"] is None
    assert (
        actual["padata"][0]["padata-value"]["cipher"] == "07404603A869C93176E28EDAD134CE7F"
        "C4C873580DF4611C855F43F6AA9E48E2"
        "F08CC28870AABCF0F7F2D4A2C2E353DE"
        "81F7302FAF7C8512"
    )
    assert actual["padata"][1]["padata-type"] == "PA-REQ-ENC-PA-REP (149)"
    assert actual["padata"][1]["padata-value"] == ""
    assert actual["req-body"]["kdc-options"]["raw"] == 1073741824
    assert actual["req-body"]["kdc-options"]["flags"] == ["forwardable (1073741824)"]
    assert actual["req-body"]["cname"]["name-type"] == "NT-PRINCIPAL (1)"
    assert actual["req-body"]["cname"]["name-string"] == ["vagrant-domain"]
    assert actual["req-body"]["realm"] == "DOMAIN.LOCAL"
    assert actual["req-body"]["sname"]["name-type"] == "NT-SRV-INST (2)"
    assert actual["req-body"]["sname"]["name-string"] == ["krbtgt", "DOMAIN.LOCAL"]
    assert actual["req-body"]["from"] is None
    assert actual["req-body"]["till"] == "2020-06-14T07:04:20+00:00"
    assert actual["req-body"]["rtime"] is None
    assert actual["req-body"]["nonce"] == 734266074
    assert actual["req-body"]["etype"] == [
        "AES256_CTS_HMAC_SHA1_96 (18)",
        "AES128_CTS_HMAC_SHA1_96 (17)",
        "DES3_CBC_SHA1 (16)",
        "RC4_HMAC (23)",
    ]
    assert actual["req-body"]["addresses"] is None
    assert actual["req-body"]["enc-authorization-data"] is None
    assert actual["req-body"]["additional-tickets"] is None


def test_unpack_krb_tgs_req():
    data = get_data("krb_tgs_req")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, kerb.KrbTgsReq)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.tgs_req
    assert isinstance(actual.padata, list)
    assert len(actual.padata) == 1
    assert actual.padata[0].data_type == kerb.KerberosPADataType.tgs_req
    assert isinstance(actual.padata[0].b_value, bytes)

    pa1_val = actual.padata[0].value
    assert pa1_val.PVNO == 5
    assert pa1_val.MESSAGE_TYPE == kerb.KerberosMessageType.ap_req
    assert pa1_val.ap_options == 0
    assert isinstance(pa1_val.authenticator.cipher, bytes)
    assert pa1_val.authenticator.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert pa1_val.authenticator.kvno is None
    assert isinstance(pa1_val.ticket.enc_part.cipher, bytes)
    assert pa1_val.ticket.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert pa1_val.ticket.enc_part.kvno == 2
    assert pa1_val.ticket.realm == b"DOMAIN.LOCAL"
    assert pa1_val.ticket.sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.srv_inst, [b"krbtgt", b"DOMAIN.LOCAL"]
    )
    assert pa1_val.ticket.tkt_vno == 5

    assert actual.req_body.additional_tickets is None
    assert actual.req_body.addresses is None
    assert actual.req_body.cname is None
    assert actual.req_body.enc_authorization_data is None
    assert actual.req_body.etype == [
        kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96,
        kerb.KerberosEncryptionType.aes128_cts_hmac_sha1_96,
        kerb.KerberosEncryptionType.des3_cbc_sha1,
        kerb.KerberosEncryptionType.rc4_hmac,
    ]
    assert actual.req_body.kdc_options == 1073807360
    assert actual.req_body.nonce == 333512069
    assert actual.req_body.postdated_from is None
    assert actual.req_body.postdated_till == datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
    assert actual.req_body.realm == b"DOMAIN.LOCAL"
    assert actual.req_body.rtime is None
    assert actual.req_body.sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.srv_hst, [b"HTTP", b"server2019.domain.local"]
    )

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "TGS-REQ (12)"
    assert isinstance(actual["padata"], list)
    assert len(actual["padata"]) == 1
    assert actual["padata"][0]["padata-type"] == "PA-TGS-REQ (1)"
    assert actual["padata"][0]["padata-value"]["pvno"] == 5
    assert actual["padata"][0]["padata-value"]["msg-type"] == "AP-REQ (14)"
    assert actual["padata"][0]["padata-value"]["ap-options"]["raw"] == 0
    assert actual["padata"][0]["padata-value"]["ap-options"]["flags"] == []
    assert actual["padata"][0]["padata-value"]["ticket"]["tkt-vno"] == 5
    assert actual["padata"][0]["padata-value"]["ticket"]["realm"] == "DOMAIN.LOCAL"
    assert actual["padata"][0]["padata-value"]["ticket"]["sname"]["name-type"] == "NT-SRV-INST (2)"
    assert actual["padata"][0]["padata-value"]["ticket"]["sname"]["name-string"] == ["krbtgt", "DOMAIN.LOCAL"]
    assert actual["padata"][0]["padata-value"]["ticket"]["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["padata"][0]["padata-value"]["ticket"]["enc-part"]["kvno"] == 2
    assert isinstance(actual["padata"][0]["padata-value"]["ticket"]["enc-part"]["cipher"], str)
    assert actual["padata"][0]["padata-value"]["authenticator"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["padata"][0]["padata-value"]["authenticator"]["kvno"] is None
    assert isinstance(actual["padata"][0]["padata-value"]["authenticator"]["cipher"], str)
    assert actual["req-body"]["kdc-options"]["raw"] == 1073807360
    assert actual["req-body"]["kdc-options"]["flags"] == ["forwardable (1073741824)", "canonicalize (65536)"]
    assert actual["req-body"]["cname"] is None
    assert actual["req-body"]["realm"] == "DOMAIN.LOCAL"
    assert actual["req-body"]["sname"]["name-type"] == "NT-SRV-HST (3)"
    assert actual["req-body"]["sname"]["name-string"] == ["HTTP", "server2019.domain.local"]
    assert actual["req-body"]["from"] is None
    assert actual["req-body"]["till"] == "1970-01-01T00:00:00+00:00"
    assert actual["req-body"]["rtime"] is None
    assert actual["req-body"]["etype"] == [
        "AES256_CTS_HMAC_SHA1_96 (18)",
        "AES128_CTS_HMAC_SHA1_96 (17)",
        "DES3_CBC_SHA1 (16)",
        "RC4_HMAC (23)",
    ]
    assert actual["req-body"]["addresses"] is None
    assert actual["req-body"]["enc-authorization-data"] is None
    assert actual["req-body"]["additional-tickets"] is None


def test_unpack_krb_as_rep():
    data = get_data("krb_as_rep")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, kerb.KrbAsRep)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.as_rep
    assert actual.cname == kerb.PrincipalName(kerb.KerberosPrincipalNameType.principal, [b"vagrant-domain"])
    assert actual.crealm == b"DOMAIN.LOCAL"

    assert isinstance(actual.enc_part.cipher, bytes)
    assert actual.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.enc_part.kvno == 11

    assert isinstance(actual.padata, list)
    assert len(actual.padata) == 1
    assert actual.padata[0].data_type == kerb.KerberosPADataType.etype_info2
    assert (
        actual.padata[0].b_value == b"\x30\x25\x30\x23\xA0\x03\x02\x01\x12\xA1\x1C\x1B\x1A\x44\x4F\x4D"
        b"\x41\x49\x4E\x2E\x4C\x4F\x43\x41\x4C\x76\x61\x67\x72\x61\x6E\x74"
        b"\x2D\x64\x6F\x6D\x61\x69\x6E"
    )
    pa1_val = actual.padata[0].value
    assert isinstance(pa1_val, list)
    assert len(pa1_val) == 1
    assert pa1_val[0].etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert pa1_val[0].salt == b"DOMAIN.LOCALvagrant-domain"
    assert pa1_val[0].s2kparams is None

    assert isinstance(actual.ticket.enc_part.cipher, bytes)
    assert actual.ticket.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.ticket.enc_part.kvno == 2
    assert actual.ticket.realm == b"DOMAIN.LOCAL"
    assert actual.ticket.sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.srv_inst, [b"krbtgt", b"DOMAIN.LOCAL"]
    )
    assert actual.ticket.tkt_vno == 5

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "AS-REP (11)"
    assert isinstance(actual["padata"], list)
    assert len(actual["padata"]) == 1
    assert actual["padata"][0]["padata-type"] == "PA-ETYPE-INFO2 (19)"
    assert isinstance(actual["padata"][0]["padata-value"], list)
    assert len(actual["padata"][0]["padata-value"]) == 1
    assert actual["padata"][0]["padata-value"][0]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["padata"][0]["padata-value"][0]["salt"] == "444F4D41494E2E4C4F43414C76616772616E742D646F6D61696E"
    assert actual["padata"][0]["padata-value"][0]["s2kparams"] is None
    assert actual["crealm"] == "DOMAIN.LOCAL"
    assert actual["cname"]["name-type"] == "NT-PRINCIPAL (1)"
    assert actual["cname"]["name-string"] == ["vagrant-domain"]
    assert actual["ticket"]["tkt-vno"] == 5
    assert actual["ticket"]["realm"] == "DOMAIN.LOCAL"
    assert actual["ticket"]["sname"]["name-type"] == "NT-SRV-INST (2)"
    assert actual["ticket"]["sname"]["name-string"] == ["krbtgt", "DOMAIN.LOCAL"]
    assert actual["ticket"]["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["ticket"]["enc-part"]["kvno"] == 2
    assert isinstance(actual["ticket"]["enc-part"]["cipher"], str)
    assert actual["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["enc-part"]["kvno"] == 11
    assert isinstance(actual["enc-part"]["cipher"], str)


def test_unpack_krb_tgs_rep():
    data = get_data("krb_tgs_rep")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, kerb.KrbTgsRep)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.tgs_rep
    assert actual.cname == kerb.PrincipalName(kerb.KerberosPrincipalNameType.principal, [b"vagrant-domain"])
    assert actual.crealm == b"DOMAIN.LOCAL"

    assert isinstance(actual.enc_part.cipher, bytes)
    assert actual.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.enc_part.kvno is None

    assert actual.padata is None

    assert isinstance(actual.ticket.enc_part.cipher, bytes)
    assert actual.ticket.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.ticket.enc_part.kvno == 6
    assert actual.ticket.realm == b"DOMAIN.LOCAL"
    assert actual.ticket.sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.srv_hst, [b"HTTP", b"server2019.domain.local"]
    )
    assert actual.ticket.tkt_vno == 5

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "TGS-REP (13)"
    assert actual["padata"] is None
    assert actual["crealm"] == "DOMAIN.LOCAL"
    assert actual["cname"]["name-type"] == "NT-PRINCIPAL (1)"
    assert actual["cname"]["name-string"] == ["vagrant-domain"]
    assert actual["ticket"]["tkt-vno"] == 5
    assert actual["ticket"]["realm"] == "DOMAIN.LOCAL"
    assert actual["ticket"]["sname"]["name-type"] == "NT-SRV-HST (3)"
    assert actual["ticket"]["sname"]["name-string"] == ["HTTP", "server2019.domain.local"]
    assert actual["ticket"]["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["ticket"]["enc-part"]["kvno"] == 6
    assert isinstance(actual["ticket"]["enc-part"]["cipher"], str)
    assert actual["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["enc-part"]["kvno"] is None
    assert isinstance(actual["enc-part"]["cipher"], str)


def test_unpack_krb_ap_req():
    data = get_data("initial_context_token_krb_ap_req")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.kerberos.value

    actual = actual.token

    assert isinstance(actual, kerb.KrbApReq)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.ap_req
    assert actual.ap_options == kerb.KerberosAPOptions.mutual_required
    assert isinstance(actual.authenticator.cipher, bytes)
    assert actual.authenticator.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.authenticator.kvno is None
    assert isinstance(actual.ticket.enc_part.cipher, bytes)
    assert actual.ticket.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.ticket.enc_part.kvno == 6
    assert actual.ticket.realm == b"DOMAIN.LOCAL"
    assert actual.ticket.sname == kerb.PrincipalName(kerb.KerberosPrincipalNameType.srv_hst, [b"host", b"dc01"])
    assert actual.ticket.tkt_vno == 5

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "AP-REQ (14)"
    assert actual["ap-options"]["raw"] == 32
    assert actual["ap-options"]["flags"] == ["mutual-required (32)"]
    assert actual["ticket"]["tkt-vno"] == 5
    assert actual["ticket"]["realm"] == "DOMAIN.LOCAL"
    assert actual["ticket"]["sname"]["name-type"] == "NT-SRV-HST (3)"
    assert actual["ticket"]["sname"]["name-string"] == ["host", "dc01"]
    assert actual["ticket"]["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["ticket"]["enc-part"]["kvno"] == 6
    assert isinstance(actual["ticket"]["enc-part"]["cipher"], str)
    assert actual["authenticator"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["authenticator"]["kvno"] is None
    assert isinstance(actual["authenticator"]["cipher"], str)


def test_unpack_krb_ap_req_unknown_options():
    data = get_data("initial_context_token_krb_ap_req")
    data = data[:40] + b"\x10" + data[41:]  # This is where ap-options is set

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.kerberos.value

    actual = actual.token

    assert isinstance(actual, kerb.KrbApReq)
    assert actual.ap_options == 16


def test_unpack_krb_ap_rep():
    data = get_data("initial_context_token_krb_ap_rep")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.kerberos.value

    actual = actual.token

    assert isinstance(actual, kerb.KrbApRep)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.ap_rep
    assert isinstance(actual.enc_part.cipher, bytes)
    assert actual.enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert actual.enc_part.kvno is None

    # Test pyspnego-parse dict.
    actual = kerb.parse_kerberos_token(actual)
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "AP-REP (15)"
    assert actual["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["enc-part"]["kvno"] is None
    assert isinstance(actual["enc-part"]["cipher"], str)


def test_unpack_krb_error():
    data = get_data("krb_error")

    actual = sp.unpack_token(data)
    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)

    assert isinstance(actual, kerb.KrbError)
    assert actual.PVNO == 5
    assert actual.MESSAGE_TYPE == kerb.KerberosMessageType.error
    assert actual.cname is None
    assert actual.crealm is None
    assert actual.ctime is None
    assert actual.cusec is None
    assert isinstance(actual.e_data, bytes)
    assert actual.e_text is None
    assert actual.error_code == kerb.KerberosErrorCode.preauth_required
    assert actual.realm == b"DOMAIN.LOCAL"
    assert actual.sname == kerb.PrincipalName(kerb.KerberosPrincipalNameType.srv_inst, [b"krbtgt", b"DOMAIN.LOCAL"])
    assert actual.stime == datetime.datetime(2020, 6, 13, 21, 4, 23, tzinfo=datetime.timezone.utc)
    assert actual.susec == 748591

    actual = kerb.parse_kerberos_token(actual, encoding="utf-8")
    assert actual["pvno"] == 5
    assert actual["msg-type"] == "KRB-ERROR (30)"
    assert actual["ctime"] is None
    assert actual["cusec"] is None
    assert actual["stime"] == "2020-06-13T21:04:23+00:00"
    assert actual["susec"] == 748591
    assert actual["error-code"] == "KDC_ERR_PREAUTH_REQUIRED (25)"
    assert actual["crealm"] is None
    assert actual["cname"] is None
    assert actual["realm"] == "DOMAIN.LOCAL"
    assert actual["sname"]["name-type"] == "NT-SRV-INST (2)"
    assert actual["sname"]["name-string"] == ["krbtgt", "DOMAIN.LOCAL"]
    assert actual["e-text"] is None
    assert isinstance(actual["e-data"], str)


def test_padata_unknown_type():
    value = b"".join(
        [
            pack_asn1(TagClass.context_specific, True, 1, pack_asn1_integer(1024)),
            pack_asn1(TagClass.context_specific, True, 2, pack_asn1_octet_string(b"")),
        ]
    )
    padata = kerb.PAData.unpack(value)

    assert padata.data_type == 1024
    assert padata.b_value == b""
    assert padata.value == b""

    actual = kerb.parse_kerberos_token(padata)

    assert isinstance(actual, dict)
    assert actual["padata-type"] == "UNKNOWN (1024)"
    assert actual["padata-value"] == ""


def test_req_body_addresses():
    value = b"".join(
        [
            pack_asn1(TagClass.context_specific, True, 0, pack_asn1_bit_string(b"\x00\x00\x00\x00")),
            pack_asn1(TagClass.context_specific, True, 2, pack_asn1_general_string(b"DOMAIN.LOCAL")),
            pack_asn1(TagClass.context_specific, True, 7, pack_asn1_integer(1)),
            pack_asn1(
                TagClass.context_specific,
                True,
                8,
                pack_asn1_sequence([pack_asn1_integer(kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96)]),
            ),
            pack_asn1(
                TagClass.context_specific,
                True,
                9,
                pack_asn1_sequence(
                    [
                        pack_asn1_sequence(
                            [
                                pack_asn1(
                                    TagClass.context_specific,
                                    True,
                                    0,
                                    pack_asn1_integer(kerb.KerberosHostAddressType.ipv4),
                                ),
                                pack_asn1(
                                    TagClass.context_specific, True, 1, pack_asn1_octet_string(b"dc01.domain.local")
                                ),
                            ]
                        )
                    ]
                ),
            ),
        ]
    )
    req_body = kerb.KdcReqBody.unpack(value)

    assert isinstance(req_body.addresses, list)
    assert len(req_body.addresses) == 1
    assert req_body.addresses[0].addr_type == kerb.KerberosHostAddressType.ipv4
    assert req_body.addresses[0].value == b"dc01.domain.local"

    actual = kerb.parse_kerberos_token(req_body)
    assert isinstance(actual, dict)
    assert actual["addresses"][0]["addr-type"] == "IPv4 (2)"
    assert actual["addresses"][0]["address"] == "dc01.domain.local"


def test_req_body_ticket():
    ticket = pack_asn1(
        TagClass.application,
        True,
        1,
        pack_asn1_sequence(
            [
                pack_asn1(TagClass.context_specific, True, 0, pack_asn1_integer(5)),
                pack_asn1(TagClass.context_specific, True, 1, pack_asn1_general_string(b"DOMAIN.LOCAL")),
                pack_asn1(
                    TagClass.context_specific,
                    True,
                    2,
                    pack_asn1_sequence(
                        [
                            pack_asn1(
                                TagClass.context_specific,
                                True,
                                0,
                                pack_asn1_integer(kerb.KerberosPrincipalNameType.principal),
                            ),
                            pack_asn1(
                                TagClass.context_specific,
                                True,
                                1,
                                pack_asn1_sequence(
                                    [
                                        pack_asn1_general_string(b"vagrant-domain"),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                pack_asn1(
                    TagClass.context_specific,
                    True,
                    3,
                    pack_asn1_sequence(
                        [
                            pack_asn1(
                                TagClass.context_specific,
                                True,
                                0,
                                pack_asn1_integer(kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96),
                            ),
                            pack_asn1(TagClass.context_specific, True, 2, pack_asn1_octet_string(b"\x00\x01")),
                        ]
                    ),
                ),
            ]
        ),
    )
    value = b"".join(
        [
            pack_asn1(TagClass.context_specific, True, 0, pack_asn1_bit_string(b"\x00\x00\x00\x00")),
            pack_asn1(TagClass.context_specific, True, 2, pack_asn1_general_string(b"DOMAIN.LOCAL")),
            pack_asn1(TagClass.context_specific, True, 7, pack_asn1_integer(1)),
            pack_asn1(
                TagClass.context_specific,
                True,
                8,
                pack_asn1_sequence([pack_asn1_integer(kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96)]),
            ),
            pack_asn1(TagClass.context_specific, True, 11, pack_asn1_sequence([ticket])),
        ]
    )
    req_body = kerb.KdcReqBody.unpack(value)

    assert isinstance(req_body.additional_tickets, list)
    assert len(req_body.additional_tickets) == 1
    assert req_body.additional_tickets[0].enc_part.cipher == b"\x00\x01"
    assert req_body.additional_tickets[0].enc_part.etype == kerb.KerberosEncryptionType.aes256_cts_hmac_sha1_96
    assert req_body.additional_tickets[0].enc_part.kvno is None
    assert req_body.additional_tickets[0].realm == b"DOMAIN.LOCAL"
    assert req_body.additional_tickets[0].sname == kerb.PrincipalName(
        kerb.KerberosPrincipalNameType.principal, [b"vagrant-domain"]
    )
    assert req_body.additional_tickets[0].tkt_vno == 5

    actual = kerb.parse_kerberos_token(req_body)
    assert isinstance(actual, dict)
    assert actual["additional-tickets"][0]["tkt-vno"] == 5
    assert actual["additional-tickets"][0]["realm"] == "DOMAIN.LOCAL"
    assert actual["additional-tickets"][0]["sname"]["name-type"] == "NT-PRINCIPAL (1)"
    assert actual["additional-tickets"][0]["sname"]["name-string"] == ["vagrant-domain"]
    assert actual["additional-tickets"][0]["enc-part"]["etype"] == "AES256_CTS_HMAC_SHA1_96 (18)"
    assert actual["additional-tickets"][0]["enc-part"]["kvno"] is None
    assert actual["additional-tickets"][0]["enc-part"]["cipher"] == "0001"



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest

import spnego
import spnego._gss
from spnego._context import GSSMech
from spnego._negotiate import NegotiateProxy
from spnego._spnego import NegState, NegTokenInit, NegTokenResp
from spnego.exceptions import BadMechanismError, InvalidTokenError


def test_token_rejected(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], options=spnego.NegotiateOptions.use_negotiate)

    c.step()
    token_resp = NegTokenResp(neg_state=NegState.reject).pack()

    with pytest.raises(InvalidTokenError, match="Received SPNEGO rejection with no token error message"):
        c.step(token_resp)


def test_token_invalid_input(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], options=spnego.NegotiateOptions.use_negotiate)

    c.step()
    with pytest.raises(InvalidTokenError, match="Failed to unpack input token"):
        c.step(b"\x00")


def test_token_no_common_mechs(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], options=spnego.NegotiateOptions.use_negotiate)

    with pytest.raises(BadMechanismError, match="Unable to negotiate common mechanism"):
        c.step(NegTokenInit(mech_types=["1.2.3.4"]).pack())


def test_token_acceptor_first(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], options=spnego.NegotiateOptions.use_negotiate)
    s = spnego.server(options=spnego.NegotiateOptions.use_negotiate)

    assert getattr(c, "_mech_list", None) == []
    assert getattr(s, "_mech_list", None) == []

    token1 = s.step()
    assert isinstance(token1, bytes)
    assert not c.complete
    assert not s.complete
    assert getattr(c, "_mech_list", None) == []
    assert GSSMech.ntlm.value in getattr(s, "_mech_list", [])

    negotiate = c.step(token1)
    assert isinstance(negotiate, bytes)
    assert not c.complete
    assert not s.complete
    assert getattr(c, "_mech_list", None) == [GSSMech.ntlm.value]
    assert GSSMech.ntlm.value in getattr(s, "_mech_list", [])

    challenge = s.step(negotiate)
    assert isinstance(challenge, bytes)
    assert not c.complete
    assert not s.complete

    authenticate = c.step(challenge)
    assert isinstance(authenticate, bytes)
    assert not c.complete
    assert not s.complete

    mech_list_mic = s.step(authenticate)
    assert isinstance(mech_list_mic, bytes)
    assert not c.complete
    assert s.complete

    final_token = c.step(mech_list_mic)
    assert final_token is None
    assert c.complete
    assert s.complete


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_iov_available_sspi():
    assert NegotiateProxy.iov_available()


@pytest.mark.skipif(not spnego._gss.HAS_GSSAPI, reason="Requires the gssapi library to be installed for testing")
def test_iov_available_gssapi():
    assert NegotiateProxy.iov_available() == spnego._gss.GSSAPIProxy.iov_available()



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest


def test_gss_import_dep():
    with pytest.deprecated_call(match="spnego.gss is deprecated and will be removed in a future release"):
        import spnego.gss


def test_negotiate_import_dep():
    with pytest.deprecated_call(match="spnego.negotiate is deprecated and will be removed in a future release"):
        import spnego.negotiate


def test_ntlm_import_dep():
    with pytest.deprecated_call(match="spnego.ntlm is deprecated and will be removed in a future release"):
        import spnego.ntlm


def test_sspi_import_dep():
    with pytest.deprecated_call(match="spnego.sspi is deprecated and will be removed in a future release"):
        import spnego.sspi



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import socket

import pytest

import spnego._sspi
import spnego.iov
from spnego.exceptions import InvalidCredentialError


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_build_iov_list(ntlm_cred):
    c = spnego._sspi.SSPIProxy(ntlm_cred[0], ntlm_cred[1], protocol="ntlm")
    c._security_trailer = 10
    c._block_size = 2

    actual = c._build_iov_list(
        [
            (spnego.iov.BufferType.header, b"\x01"),
            (spnego.iov.BufferType.data, 5),
            (spnego.iov.BufferType.padding, True),
            spnego.iov.BufferType.header,
            spnego.iov.BufferType.stream,
            b"\x02",
        ],
        c._convert_iov_buffer,
    )

    assert len(actual) == 6
    assert actual[0].buffer_type == spnego.iov.BufferType.header
    assert actual[0].data == b"\x01"
    assert actual[1].buffer_type == spnego.iov.BufferType.data
    assert actual[1].data == b"\x00" * 5
    assert actual[2].buffer_type == spnego.iov.BufferType.padding
    assert actual[2].data == b"\x00" * 2
    assert actual[3].buffer_type == spnego.iov.BufferType.header
    assert actual[3].data == b"\x00" * 10
    assert actual[4].buffer_type == spnego.iov.BufferType.stream
    assert actual[4].data == b""
    assert actual[5].buffer_type == spnego.iov.BufferType.data
    assert actual[5].data == b"\x02"


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_build_iov_list_fail_auto_alloc(ntlm_cred):
    c = spnego._sspi.SSPIProxy(ntlm_cred[0], ntlm_cred[1], protocol="ntlm")
    c._security_trailer = 10
    c._block_size = 2

    with pytest.raises(ValueError, match="Cannot auto allocate buffer of type BufferType.data"):
        c._build_iov_list([(spnego.iov.BufferType.data, True)], c._convert_iov_buffer)


def test_no_sspi_library(monkeypatch):
    monkeypatch.setattr(spnego._sspi, "HAS_SSPI", False)

    with pytest.raises(ImportError, match="SSPIProxy requires the Windows only sspilib python package"):
        spnego._sspi.SSPIProxy()


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_sspi_invalid_qop():
    c = spnego._sspi.SSPIProxy("user", "pass")

    with pytest.raises(ValueError, match="Cannot set qop with SECQOP_WRAP_NO_ENCRYPT and encrypt=True"):
        c.wrap(b"\x00", True, qop=0x80000001)


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_sspi_wrap_no_encryption(ntlm_cred):
    c = spnego._sspi.SSPIProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname())
    s = spnego._sspi.SSPIProxy(usage="accept")

    s.step(c.step(s.step(c.step())))

    plaintext = os.urandom(16)
    enc_data = c.wrap(plaintext, encrypt=False)
    dec_data = s.unwrap(enc_data.data)
    assert dec_data.data == plaintext


@pytest.mark.skipif("ntlm" not in spnego._sspi.SSPIProxy.available_protocols(), reason="Requires SSPI library")
def test_sspi_no_valid_cred():
    with pytest.raises(InvalidCredentialError, match="No applicable credentials available"):
        spnego._sspi.SSPIProxy(spnego.KerberosKeytab("user_princ", "ccache"), protocol="kerberos")



import hashlib
import os
import random

import pytest

from spnego._ntlm_raw.md4 import md4

HASHLIB_AVAIL = True
try:
    hashlib.new("md4")
except ValueError:
    HASHLIB_AVAIL = False


# The expectations here are based on the MD4 RFC
# https://datatracker.ietf.org/doc/html/rfc1320#appendix-A.5
@pytest.mark.parametrize(
    "data, expected",
    [
        (
            b"",
            b"\x31\xd6\xcf\xe0\xd1\x6a\xe9\x31\xb7\x3c\x59\xd7\xe0\xc0\x89\xc0",
        ),
        (
            b"a",
            b"\xbd\xe5\x2c\xb3\x1d\xe3\x3e\x46\x24\x5e\x05\xfb\xdb\xd6\xfb\x24",
        ),
        (
            b"abc",
            b"\xa4\x48\x01\x7a\xaf\x21\xd8\x52\x5f\xc1\x0a\xe8\x7a\xa6\x72\x9d",
        ),
        (
            b"message digest",
            b"\xd9\x13\x0a\x81\x64\x54\x9f\xe8\x18\x87\x48\x06\xe1\xc7\x01\x4b",
        ),
        (
            b"abcdefghijklmnopqrstuvwxyz",
            b"\xd7\x9e\x1c\x30\x8a\xa5\xbb\xcd\xee\xa8\xed\x63\xdf\x41\x2d\xa9",
        ),
        (
            b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
            b"\x04\x3f\x85\x82\xf2\x41\xdb\x35\x1c\xe6\x27\xe1\x53\xe7\xf0\xe4",
        ),
        (
            b"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
            b"\xe3\x3b\x4d\xdc\x9c\x38\xf2\x19\x9c\x3e\x7b\x16\x4f\xcc\x05\x36",
        ),
    ],
)
def test_md4(data: bytes, expected: bytes) -> None:
    actual = md4(data)
    assert actual == expected


# This test will only run on hosts where md4 is still available on hashlib.
# It's just an extra sanity check to verify our implementation is still good.
@pytest.mark.skipif(not HASHLIB_AVAIL, reason="hashlib does not support md4")
def test_md4_to_hashlib() -> None:
    for idx in range(20):
        data = os.urandom(random.randint(idx * 10, (idx * 10) + 1024))

        expected = hashlib.new("md4", data).digest()
        actual = md4(data)

        assert actual == expected



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import re
import typing

import pytest

import spnego
import spnego._gss
import spnego.iov
from spnego.exceptions import InvalidCredentialError, NoContextError


def test_gss_sasl_description_fail(mocker, monkeypatch):
    gssapi = pytest.importorskip("gssapi")
    SASLResult = collections.namedtuple("SASLResult", ["mech_description"])

    mock_inquire_sasl = mocker.MagicMock()
    mock_inquire_sasl.side_effect = [Exception, SASLResult(b"result")]
    monkeypatch.setattr(gssapi.raw, "inquire_saslname_for_mech", mock_inquire_sasl)

    actual = spnego._gss._gss_sasl_description(gssapi.OID.from_int_seq("1.2.3"))
    assert actual is None

    actual = spnego._gss._gss_sasl_description(gssapi.OID.from_int_seq("1.2.3"))
    assert actual is None

    actual = spnego._gss._gss_sasl_description(gssapi.OID.from_int_seq("1.2.3.4"))
    assert actual == b"result"

    actual = spnego._gss._gss_sasl_description(gssapi.OID.from_int_seq("1.2.3.4"))
    assert actual == b"result"

    assert mock_inquire_sasl.call_count == 2


def test_build_iov_list(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")
    actual = c._build_iov_list(
        [
            (spnego.iov.BufferType.header, b"\x01"),
            (spnego.iov.BufferType.data, 1),
            (spnego.iov.BufferType.padding, True),
            spnego.iov.BufferType.header,
            spnego.iov.BufferType.stream,
            b"\x02",
        ],
        c._convert_iov_buffer,
    )

    assert len(actual) == 6
    assert actual[0] == (spnego.iov.BufferType.header, False, b"\x01")
    assert actual[1] == (spnego.iov.BufferType.data, False, b"\x00")
    assert actual[2] == (spnego.iov.BufferType.padding, True, None)
    assert actual[3] == (spnego.iov.BufferType.header, True, None)
    assert actual[4] == (spnego.iov.BufferType.stream, False, None)
    assert actual[5] == (spnego.iov.BufferType.data, False, b"\x02")


def test_gssapi_query_message_sizes_fail(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot get message sizes until context has been established"):
        c.query_message_sizes()


def test_gssapi_wrap_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        c.wrap(b"data")


def test_gssapi_wrap_iov_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        c.wrap_iov([])


def test_gssapi_wrap_winrm_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        c.wrap_winrm(b"data")


def test_gssapi_unwrap_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        c.unwrap(b"data")


def test_gssapi_unwrap_iov_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        c.unwrap_iov([])


def test_gssapi_unwrap_winrm_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        c.unwrap_winrm(b"header", b"data")


def test_gssapi_sign_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot sign until context has been established"):
        c.sign(b"data")


def test_gssapi_verify_no_context(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    with pytest.raises(NoContextError, match="Cannot verify until context has been established"):
        c.verify(b"data", b"mic")


def test_build_iov_list_invalid_tuple(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    expected = "IOV entry tuple must contain 2 values, the type and data, see IOVBuffer."
    with pytest.raises(ValueError, match=expected):
        c._build_iov_list([(1, 2, 3)], c._convert_iov_buffer)  # type: ignore[list-item] # we are testing this


def test_build_iov_list_invalid_buffer_type(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    expected = "IOV entry[0] must specify the BufferType as an int"
    with pytest.raises(ValueError, match=re.escape(expected)):
        c._build_iov_list([(b"", b"")], c._convert_iov_buffer)  # type: ignore[list-item] # we are testing this


def test_build_iov_list_invalid_data(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    expected = "IOV entry[1] must specify the buffer bytes, length of the buffer, or whether it is auto allocated."
    with pytest.raises(ValueError, match=re.escape(expected)):
        c._build_iov_list([(1, "data")], c._convert_iov_buffer)  # type: ignore[list-item] # we are testing this


def test_build_iov_list_invalid_value(kerb_cred):
    c = spnego._gss.GSSAPIProxy(kerb_cred.user_princ, protocol="kerberos")

    expected = "IOV entry must be a IOVBuffer tuple, int, or bytes"
    with pytest.raises(ValueError, match=re.escape(expected)):
        c._build_iov_list([None], c._convert_iov_buffer)  # type: ignore[list-item] # we are testing this


def test_no_gssapi_library(monkeypatch):
    monkeypatch.setattr(spnego._gss, "HAS_GSSAPI", False)

    with pytest.raises(ImportError, match="GSSAPIProxy requires the Python gssapi library"):
        spnego._gss.GSSAPIProxy()


@pytest.mark.skipif(not spnego._gss.HAS_GSSAPI, reason="Requires the gssapi library to be installed for testing")
def test_gssapi_no_valid_acceptor_cred():
    server_creds: typing.List[typing.Tuple[spnego.Credential, str]] = [
        (spnego.KerberosCCache("ccache"), "kerberos"),
        (spnego.KerberosKeytab("user_princ", "ccache"), "kerberos"),
        (spnego.NTLMHash("user_princ"), "ntlm"),
    ]
    for cred, protocol in server_creds:
        if protocol not in spnego._gss.GSSAPIProxy.available_protocols():
            continue

        with pytest.raises(InvalidCredentialError, match="No applicable credentials available"):
            spnego._gss.GSSAPIProxy(cred, protocol=protocol, usage="accept")


@pytest.mark.parametrize(
    "value, expected",
    [
        ("", b""),
        ("foo", b"foo"),
        ("café", b"caf\xC3\xA9"),
        (
            b"\xDD\xBA\xE2\xD9\x12\x53".decode("utf-16-le", errors="surrogatepass"),
            b"\xEB\xAB\x9D\xEF\xBF\xBD\xE5\x8C\x92",
        ),
        (
            b"\xDD\xBA\xE2\xD9\x12\x53".decode("utf-16-le", errors="replace"),
            b"\xEB\xAB\x9D\xEF\xBF\xBD\xE5\x8C\x92",
        ),
    ],
)
def test_encode_password(
    value: str,
    expected: bytes,
) -> None:
    actual = spnego._gss._encode_kerb_password(value)
    assert actual == expected
